from io import BytesIO
import textwrap
import email
import time
import random
import threading
import json
import asyncio
import ssl
from http.client import HTTPResponse

# Used for global variables
import stripe  # noqa: IMP101
from stripe import _util
from stripe._request_metrics import RequestMetrics
from stripe._error import APIConnectionError

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    ClassVar,
    Union,
    cast,
    overload,
    AsyncIterable,
)
from typing_extensions import (
    TYPE_CHECKING,
    Literal,
    NoReturn,
    TypedDict,
    Awaitable,
    Never,
)

if TYPE_CHECKING:
    from urllib.parse import ParseResult

    try:
        from requests import Session as RequestsSession
    except ImportError:
        pass

    try:
        from httpx import Timeout as HTTPXTimeout
        from httpx import Client as HTTPXClientType
    except ImportError:
        pass

    try:
        from aiohttp import ClientTimeout as AIOHTTPTimeout
        from aiohttp import StreamReader as AIOHTTPStreamReader
    except ImportError:
        pass


def _now_ms():
    return int(round(time.time() * 1000))


def new_default_http_client(*args: Any, **kwargs: Any) -> "HTTPClient":
    return _default_sync_client(*args, **kwargs)


def new_http_client_async_fallback(*args: Any, **kwargs: Any) -> "HTTPClient":
    return _default_async_client(*args, **kwargs)


class HTTPClient(object):
    """
    Base HTTP client that custom clients can inherit from.
    """

    name: ClassVar[str]

    class _Proxy(TypedDict):
        http: Optional[str]
        https: Optional[str]

    MAX_DELAY = 5
    INITIAL_DELAY = 0.5
    MAX_RETRY_AFTER = 60
    _proxy: Optional[_Proxy]
    _verify_ssl_certs: bool

    def __init__(
        self,
        verify_ssl_certs: bool = True,
        proxy: Optional[Union[str, _Proxy]] = None,
        async_fallback_client: Optional["HTTPClient"] = None,
        _lib=None,  # used for internal unit testing
    ):
        self._verify_ssl_certs = verify_ssl_certs
        if proxy:
            if isinstance(proxy, str):
                proxy = HTTPClient._Proxy(http=proxy, https=proxy)
            if not isinstance(proxy, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise ValueError(
                    "Proxy(ies) must be specified as either a string "
                    "URL or a dict() with string URL under the"
                    " "
                    "https"
                    " and/or "
                    "http"
                    " keys."
                )
        self._proxy = proxy.copy() if proxy else None
        self._async_fallback_client = async_fallback_client

        self._thread_local = threading.local()

    def _should_retry(
        self,
        response: Optional[Tuple[Any, int, Optional[Mapping[str, str]]]],
        api_connection_error: Optional[APIConnectionError],
        num_retries: int,
        max_network_retries: Optional[int],
    ):
        max_network_retries = (
            max_network_retries if max_network_retries is not None else 0
        )
        if num_retries >= max_network_retries:
            return False

        if response is None:
            # We generally want to retry on timeout and connection
            # exceptions, but defer this decision to underlying subclass
            # implementations. They should evaluate the driver-specific
            # errors worthy of retries, and set flag on the error returned.
            assert api_connection_error is not None
            return api_connection_error.should_retry

        _, status_code, rheaders = response

        # The API may ask us not to retry (eg; if doing so would be a no-op)
        # or advise us to retry (eg; in cases of lock timeouts); we defer to that.
        #
        # Note that we expect the headers object to be a CaseInsensitiveDict, as is the case with the requests library.
        if rheaders is not None and "stripe-should-retry" in rheaders:
            if rheaders["stripe-should-retry"] == "false":
                return False
            if rheaders["stripe-should-retry"] == "true":
                return True

        # Retry on conflict errors.
        if status_code == 409:
            return True

        # Retry on 500, 503, and other internal errors.
        #
        # Note that we expect the stripe-should-retry header to be false
        # in most cases when a 500 is returned, since our idempotency framework
        # would typically replay it anyway.
        if status_code >= 500:
            return True

        return False

    def _retry_after_header(
        self, response: Optional[Tuple[Any, Any, Mapping[str, str]]] = None
    ):
        if response is None:
            return None
        _, _, rheaders = response

        try:
            return int(rheaders["retry-after"])
        except (KeyError, ValueError):
            return None

    def _sleep_time_seconds(
        self,
        num_retries: int,
        response: Optional[Tuple[Any, Any, Mapping[str, str]]] = None,
    ) -> float:
        """
        Apply exponential backoff with initial_network_retry_delay on the number of num_retries so far as inputs.
        Do not allow the number to exceed `max_network_retry_delay`.
        """
        sleep_seconds = min(
            HTTPClient.INITIAL_DELAY * (2 ** (num_retries - 1)),
            HTTPClient.MAX_DELAY,
        )

        sleep_seconds = self._add_jitter_time(sleep_seconds)

        # But never sleep less than the base sleep seconds.
        sleep_seconds = max(HTTPClient.INITIAL_DELAY, sleep_seconds)

        # And never sleep less than the time the API asks us to wait, assuming it's a reasonable ask.
        retry_after = self._retry_after_header(response) or 0
        if retry_after <= HTTPClient.MAX_RETRY_AFTER:
            sleep_seconds = max(retry_after, sleep_seconds)

        return sleep_seconds

    def _add_jitter_time(self, sleep_seconds: float) -> float:
        """
        Randomize the value in `[(sleep_seconds/ 2) to (sleep_seconds)]`.
        Also separated method here to isolate randomness for tests
        """
        sleep_seconds *= 0.5 * (1 + random.uniform(0, 1))
        return sleep_seconds

    def _add_telemetry_header(
        self, headers: Mapping[str, str]
    ) -> Mapping[str, str]:
        last_request_metrics = getattr(
            self._thread_local, "last_request_metrics", None
        )
        if stripe.enable_telemetry and last_request_metrics:
            telemetry = {
                "last_request_metrics": last_request_metrics.payload()
            }
            ret = dict(headers)
            ret["X-Stripe-Client-Telemetry"] = json.dumps(telemetry)
            return ret
        return headers

    def _record_request_metrics(self, response, request_start, usage):
        _, _, rheaders = response
        if "Request-Id" in rheaders and stripe.enable_telemetry:
            request_id = rheaders["Request-Id"]
            request_duration_ms = _now_ms() - request_start
            self._thread_local.last_request_metrics = RequestMetrics(
                request_id, request_duration_ms, usage=usage
            )

    def request_with_retries(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data: Any = None,
        max_network_retries: Optional[int] = None,
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[str, int, Mapping[str, str]]:
        return self._request_with_retries_internal(
            method,
            url,
            headers,
            post_data,
            is_streaming=False,
            max_network_retries=max_network_retries,
            _usage=_usage,
        )

    def request_stream_with_retries(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data=None,
        max_network_retries=None,
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[Any, int, Mapping[str, str]]:
        return self._request_with_retries_internal(
            method,
            url,
            headers,
            post_data,
            is_streaming=True,
            max_network_retries=max_network_retries,
            _usage=_usage,
        )

    def _request_with_retries_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data: Any,
        is_streaming: bool,
        max_network_retries: Optional[int],
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[Any, int, Mapping[str, str]]:
        headers = self._add_telemetry_header(headers)

        num_retries = 0

        while True:
            request_start = _now_ms()

            try:
                if is_streaming:
                    response = self.request_stream(
                        method, url, headers, post_data
                    )
                else:
                    response = self.request(method, url, headers, post_data)
                connection_error = None
            except APIConnectionError as e:
                connection_error = e
                response = None

            if self._should_retry(
                response, connection_error, num_retries, max_network_retries
            ):
                if connection_error:
                    _util.log_info(
                        "Encountered a retryable error %s"
                        % connection_error.user_message
                    )
                num_retries += 1
                sleep_time = self._sleep_time_seconds(num_retries, response)
                _util.log_info(
                    (
                        "Initiating retry %i for request %s %s after "
                        "sleeping %.2f seconds."
                        % (num_retries, method, url, sleep_time)
                    )
                )
                time.sleep(sleep_time)
            else:
                if response is not None:
                    self._record_request_metrics(
                        response, request_start, usage=_usage
                    )

                    return response
                else:
                    assert connection_error is not None
                    raise connection_error

    def request(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data: Any = None,
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[str, int, Mapping[str, str]]:
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request`"
        )

    def request_stream(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data: Any = None,
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[Any, int, Mapping[str, str]]:
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request_stream`"
        )

    def close(self):
        raise NotImplementedError(
            "HTTPClient subclasses must implement `close`"
        )

    async def request_with_retries_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data=None,
        max_network_retries: Optional[int] = None,
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[Any, int, Any]:
        return await self._request_with_retries_internal_async(
            method,
            url,
            headers,
            post_data,
            is_streaming=False,
            max_network_retries=max_network_retries,
            _usage=_usage,
        )

    async def request_stream_with_retries_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data=None,
        max_network_retries=None,
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[AsyncIterable[bytes], int, Any]:
        return await self._request_with_retries_internal_async(
            method,
            url,
            headers,
            post_data,
            is_streaming=True,
            max_network_retries=max_network_retries,
            _usage=_usage,
        )

    @overload
    async def _request_with_retries_internal_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[False],
        max_network_retries: Optional[int],
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[Any, int, Mapping[str, str]]: ...

    @overload
    async def _request_with_retries_internal_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[True],
        max_network_retries: Optional[int],
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[AsyncIterable[bytes], int, Mapping[str, str]]: ...

    async def _request_with_retries_internal_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: bool,
        max_network_retries: Optional[int],
        *,
        _usage: Optional[List[str]] = None,
    ) -> Tuple[Any, int, Mapping[str, str]]:
        headers = self._add_telemetry_header(headers)

        num_retries = 0

        while True:
            request_start = _now_ms()

            try:
                if is_streaming:
                    response = await self.request_stream_async(
                        method, url, headers, post_data
                    )
                else:
                    response = await self.request_async(
                        method, url, headers, post_data
                    )
                connection_error = None
            except APIConnectionError as e:
                connection_error = e
                response = None

            if self._should_retry(
                response, connection_error, num_retries, max_network_retries
            ):
                if connection_error:
                    _util.log_info(
                        "Encountered a retryable error %s"
                        % connection_error.user_message
                    )
                num_retries += 1
                sleep_time = self._sleep_time_seconds(num_retries, response)
                _util.log_info(
                    (
                        "Initiating retry %i for request %s %s after "
                        "sleeping %.2f seconds."
                        % (num_retries, method, url, sleep_time)
                    )
                )
                await self.sleep_async(sleep_time)
            else:
                if response is not None:
                    self._record_request_metrics(
                        response, request_start, usage=_usage
                    )

                    return response
                else:
                    assert connection_error is not None
                    raise connection_error

    async def request_async(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[bytes, int, Mapping[str, str]]:
        if self._async_fallback_client is not None:
            return await self._async_fallback_client.request_async(
                method, url, headers, post_data
            )
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request_async`"
        )

    async def request_stream_async(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[AsyncIterable[bytes], int, Mapping[str, str]]:
        if self._async_fallback_client is not None:
            return await self._async_fallback_client.request_stream_async(
                method, url, headers, post_data
            )
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request_stream_async`"
        )

    async def close_async(self):
        if self._async_fallback_client is not None:
            return await self._async_fallback_client.close_async()
        raise NotImplementedError(
            "HTTPClient subclasses must implement `close_async`"
        )

    def sleep_async(self, secs: float) -> Awaitable[None]:
        if self._async_fallback_client is not None:
            return self._async_fallback_client.sleep_async(secs)
        raise NotImplementedError(
            "HTTPClient subclasses must implement `sleep`"
        )


class RequestsClient(HTTPClient):
    name = "requests"

    def __init__(
        self,
        timeout: Union[float, Tuple[float, float]] = 80,
        session: Optional["RequestsSession"] = None,
        verify_ssl_certs: bool = True,
        proxy: Optional[Union[str, HTTPClient._Proxy]] = None,
        async_fallback_client: Optional[HTTPClient] = None,
        _lib=None,  # used for internal unit testing
        **kwargs,
    ):
        super(RequestsClient, self).__init__(
            verify_ssl_certs=verify_ssl_certs,
            proxy=proxy,
            async_fallback_client=async_fallback_client,
        )
        self._session = session
        self._timeout = timeout

        if _lib is None:
            import requests

            _lib = requests

        self.requests = _lib

    def request(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data=None,
    ) -> Tuple[bytes, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=False
        )

    def request_stream(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data=None,
    ) -> Tuple[Any, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=True
        )

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data,
        is_streaming: Literal[True],
    ) -> Tuple[Any, int, Mapping[str, str]]: ...

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data,
        is_streaming: Literal[False],
    ) -> Tuple[bytes, int, Mapping[str, str]]: ...

    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Optional[Mapping[str, str]],
        post_data,
        is_streaming: bool,
    ) -> Tuple[Union[bytes, Any], int, Mapping[str, str]]:
        kwargs = {}
        if self._verify_ssl_certs:
            kwargs["verify"] = stripe.ca_bundle_path
        else:
            kwargs["verify"] = False

        if self._proxy:
            kwargs["proxies"] = self._proxy

        if is_streaming:
            kwargs["stream"] = True

        if getattr(self._thread_local, "session", None) is None:
            self._thread_local.session = (
                self._session or self.requests.Session()
            )

        try:
            try:
                result = cast(
                    "RequestsSession", self._thread_local.session
                ).request(
                    method,
                    url,
                    headers=headers,
                    data=post_data,
                    timeout=self._timeout,
                    **kwargs,
                )
            except TypeError as e:
                raise TypeError(
                    "Warning: It looks like your installed version of the "
                    '"requests" library is not compatible with Stripe\'s '
                    "usage thereof. (HINT: The most likely cause is that "
                    'your "requests" library is out of date. You can fix '
                    'that by running "pip install -U requests".) The '
                    "underlying error was: %s" % (e,)
                )

            if is_streaming:
                content = result.raw
            else:
                # This causes the content to actually be read, which could cause
                # e.g. a socket timeout. TODO: The other fetch methods probably
                # are susceptible to the same and should be updated.
                content = result.content

            status_code = result.status_code
        except Exception as e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self._handle_request_error(e)

        return content, status_code, result.headers

    def _handle_request_error(self, e: Exception) -> NoReturn:
        # Catch SSL error first as it belongs to ConnectionError,
        # but we don't want to retry
        if isinstance(e, self.requests.exceptions.SSLError):
            msg = (
                "Could not verify Stripe's SSL certificate.  Please make "
                "sure that your network is not intercepting certificates.  "
                "If this problem persists, let us know at "
                "support@stripe.com."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        # Retry only timeout and connect errors; similar to urllib3 Retry
        elif isinstance(
            e,
            (
                self.requests.exceptions.Timeout,
                self.requests.exceptions.ConnectionError,
            ),
        ):
            msg = (
                "Unexpected error communicating with Stripe.  "
                "If this problem persists, let us know at "
                "support@stripe.com."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = True
        # Catch remaining request exceptions
        elif isinstance(e, self.requests.exceptions.RequestException):
            msg = (
                "Unexpected error communicating with Stripe.  "
                "If this problem persists, let us know at "
                "support@stripe.com."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        else:
            msg = (
                "Unexpected error communicating with Stripe. "
                "It looks like there's probably a configuration "
                "issue locally.  If this problem persists, let us "
                "know at support@stripe.com."
            )
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
            should_retry = False

        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise APIConnectionError(msg, should_retry=should_retry) from e

    def close(self):
        if getattr(self._thread_local, "session", None) is not None:
            self._thread_local.session.close()


class UrlFetchClient(HTTPClient):
    name = "urlfetch"

    def __init__(
        self,
        verify_ssl_certs: bool = True,
        proxy: Optional[HTTPClient._Proxy] = None,
        deadline: int = 55,
        async_fallback_client: Optional[HTTPClient] = None,
        _lib=None,  # used for internal unit testing
    ):
        super(UrlFetchClient, self).__init__(
            verify_ssl_certs=verify_ssl_certs,
            proxy=proxy,
            async_fallback_client=async_fallback_client,
        )

        # no proxy support in urlfetch. for a patch, see:
        # https://code.google.com/p/googleappengine/issues/detail?id=544
        if proxy:
            raise ValueError(
                "No proxy support in urlfetch library. "
                "Set stripe.default_http_client to either RequestsClient, "
                "PycurlClient, or UrllibClient instance to use a proxy."
            )

        self._verify_ssl_certs = verify_ssl_certs
        # GAE requests time out after 60 seconds, so make sure to default
        # to 55 seconds to allow for a slow Stripe
        self._deadline = deadline

        if _lib is None:
            from google.appengine.api import urlfetch  # pyright: ignore

            _lib = urlfetch

        self.urlfetch = _lib

    def request(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[str, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=False
        )

    def request_stream(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[BytesIO, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=True
        )

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[True],
    ) -> Tuple[BytesIO, int, Any]: ...

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[False],
    ) -> Tuple[str, int, Any]: ...

    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming,
    ):
        try:
            result = self.urlfetch.fetch(
                url=url,
                method=method,
                headers=headers,
                # Google App Engine doesn't let us specify our own cert bundle.
                # However, that's ok because the CA bundle they use recognizes
                # api.stripe.com.
                validate_certificate=self._verify_ssl_certs,
                deadline=self._deadline,
                payload=post_data,
            )
        except self.urlfetch.Error as e:
            self._handle_request_error(e, url)

        if is_streaming:
            # This doesn't really stream.
            content = BytesIO(str.encode(result.content))
        else:
            content = result.content

        return content, result.status_code, result.headers

    def _handle_request_error(self, e: Exception, url: str) -> NoReturn:
        if isinstance(e, self.urlfetch.InvalidURLError):
            msg = (
                "The Stripe library attempted to fetch an "
                "invalid URL (%r). This is likely due to a bug "
                "in the Stripe Python bindings. Please let us know "
                "at support@stripe.com." % (url,)
            )
        elif isinstance(e, self.urlfetch.DownloadError):
            msg = "There was a problem retrieving data from Stripe."
        elif isinstance(e, self.urlfetch.ResponseTooLargeError):
            msg = (
                "There was a problem receiving all of your data from "
                "Stripe.  This is likely due to a bug in Stripe. "
                "Please let us know at support@stripe.com."
            )
        else:
            msg = (
                "Unexpected error communicating with Stripe. If this "
                "problem persists, let us know at support@stripe.com."
            )

        msg = textwrap.fill(msg) + "\n\n(Network error: " + str(e) + ")"
        raise APIConnectionError(msg) from e

    def close(self):
        pass


class PycurlClient(HTTPClient):
    class _ParsedProxy(TypedDict, total=False):
        http: Optional["ParseResult"]
        https: Optional["ParseResult"]

    name = "pycurl"
    _parsed_proxy: Optional[_ParsedProxy]

    def __init__(
        self,
        verify_ssl_certs: bool = True,
        proxy: Optional[HTTPClient._Proxy] = None,
        async_fallback_client: Optional[HTTPClient] = None,
        _lib=None,  # used for internal unit testing
    ):
        super(PycurlClient, self).__init__(
            verify_ssl_certs=verify_ssl_certs,
            proxy=proxy,
            async_fallback_client=async_fallback_client,
        )

        if _lib is None:
            import pycurl  # pyright: ignore[reportMissingModuleSource]

            _lib = pycurl

        self.pycurl = _lib
        # Initialize this within the object so that we can reuse connections.
        self._curl = _lib.Curl()

        self._parsed_proxy = {}
        # need to urlparse the proxy, since PyCurl
        # consumes the proxy url in small pieces
        if self._proxy:
            from urllib.parse import urlparse

            proxy_ = self._proxy
            for scheme, value in proxy_.items():
                # In general, TypedDict.items() gives you (key: str, value: object)
                # but we know value to be a string because all the value types on Proxy_ are strings.
                self._parsed_proxy[scheme] = urlparse(cast(str, value))

    def parse_headers(self, data):
        if "\r\n" not in data:
            return {}
        raw_headers = data.split("\r\n", 1)[1]
        headers = email.message_from_string(raw_headers)
        return dict((k.lower(), v) for k, v in dict(headers).items())

    def request(
        self, method, url, headers: Mapping[str, str], post_data=None
    ) -> Tuple[str, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=False
        )

    def request_stream(
        self, method, url, headers: Mapping[str, str], post_data=None
    ) -> Tuple[BytesIO, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=True
        )

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[True],
    ) -> Tuple[BytesIO, int, Any]: ...

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[False],
    ) -> Tuple[str, int, Mapping[str, str]]: ...

    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming,
    ) -> Tuple[Union[str, BytesIO], int, Mapping[str, str]]:
        b = BytesIO()
        rheaders = BytesIO()

        # Pycurl's design is a little weird: although we set per-request
        # options on this object, it's also capable of maintaining established
        # connections. Here we call reset() between uses to make sure it's in a
        # pristine state, but notably reset() doesn't reset connections, so we
        # still get to take advantage of those by virtue of re-using the same
        # object.
        self._curl.reset()

        proxy = self._get_proxy(url)
        if proxy:
            if proxy.hostname:
                self._curl.setopt(self.pycurl.PROXY, proxy.hostname)
            if proxy.port:
                self._curl.setopt(self.pycurl.PROXYPORT, proxy.port)
            if proxy.username or proxy.password:
                self._curl.setopt(
                    self.pycurl.PROXYUSERPWD,
                    "%s:%s" % (proxy.username, proxy.password),
                )

        if method == "get":
            self._curl.setopt(self.pycurl.HTTPGET, 1)
        elif method == "post":
            self._curl.setopt(self.pycurl.POST, 1)
            self._curl.setopt(self.pycurl.POSTFIELDS, post_data)
        else:
            self._curl.setopt(self.pycurl.CUSTOMREQUEST, method.upper())

        # pycurl doesn't like unicode URLs
        self._curl.setopt(self.pycurl.URL, url)

        self._curl.setopt(self.pycurl.WRITEFUNCTION, b.write)
        self._curl.setopt(self.pycurl.HEADERFUNCTION, rheaders.write)
        self._curl.setopt(self.pycurl.NOSIGNAL, 1)
        self._curl.setopt(self.pycurl.CONNECTTIMEOUT, 30)
        self._curl.setopt(self.pycurl.TIMEOUT, 80)
        self._curl.setopt(
            self.pycurl.HTTPHEADER,
            ["%s: %s" % (k, v) for k, v in dict(headers).items()],
        )
        if self._verify_ssl_certs:
            self._curl.setopt(self.pycurl.CAINFO, stripe.ca_bundle_path)
        else:
            self._curl.setopt(self.pycurl.SSL_VERIFYHOST, False)

        try:
            self._curl.perform()
        except self.pycurl.error as e:
            self._handle_request_error(e)

        if is_streaming:
            b.seek(0)
            rcontent = b
        else:
            rcontent = b.getvalue().decode("utf-8")

        rcode = self._curl.getinfo(self.pycurl.RESPONSE_CODE)
        headers = self.parse_headers(rheaders.getvalue().decode("utf-8"))

        return rcontent, rcode, headers

    def _handle_request_error(self, e: Exception) -> NoReturn:
        if e.args[0] in [
            self.pycurl.E_COULDNT_CONNECT,
            self.pycurl.E_COULDNT_RESOLVE_HOST,
            self.pycurl.E_OPERATION_TIMEOUTED,
        ]:
            msg = (
                "Could not connect to Stripe.  Please check your "
                "internet connection and try again.  If this problem "
                "persists, you should check Stripe's service status at "
                "https://twitter.com/stripestatus, or let us know at "
                "support@stripe.com."
            )
            should_retry = True
        elif e.args[0] in [
            self.pycurl.E_SSL_CACERT,
            self.pycurl.E_SSL_PEER_CERTIFICATE,
        ]:
            msg = (
                "Could not verify Stripe's SSL certificate.  Please make "
                "sure that your network is not intercepting certificates.  "
                "If this problem persists, let us know at "
                "support@stripe.com."
            )
            should_retry = False
        else:
            msg = (
                "Unexpected error communicating with Stripe. If this "
                "problem persists, let us know at support@stripe.com."
            )
            should_retry = False

        msg = textwrap.fill(msg) + "\n\n(Network error: " + e.args[1] + ")"
        raise APIConnectionError(msg, should_retry=should_retry) from e

    def _get_proxy(self, url) -> Optional["ParseResult"]:
        if self._parsed_proxy:
            proxy = self._parsed_proxy
            scheme = url.split(":")[0] if url else None
            if scheme:
                return proxy.get(scheme, proxy.get(scheme[0:-1]))
        return None

    def close(self):
        pass


class UrllibClient(HTTPClient):
    name = "urllib.request"

    def __init__(
        self,
        verify_ssl_certs: bool = True,
        proxy: Optional[HTTPClient._Proxy] = None,
        async_fallback_client: Optional[HTTPClient] = None,
        _lib=None,  # used for internal unit testing
    ):
        super(UrllibClient, self).__init__(
            verify_ssl_certs=verify_ssl_certs,
            proxy=proxy,
            async_fallback_client=async_fallback_client,
        )

        if _lib is None:
            import urllib.request as urllibrequest

            _lib = urllibrequest
        self.urllibrequest = _lib

        import urllib.error as urlliberror

        self.urlliberror = urlliberror

        # prepare and cache proxy tied opener here
        self._opener = None
        if self._proxy:
            # We have to cast _Proxy to Dict[str, str] because pyright is not smart enough to
            # realize that all the value types are str.
            proxy_handler = self.urllibrequest.ProxyHandler(
                cast(Dict[str, str], self._proxy)
            )
            self._opener = self.urllibrequest.build_opener(proxy_handler)

    def request(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[str, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=False
        )

    def request_stream(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[HTTPResponse, int, Mapping[str, str]]:
        return self._request_internal(
            method, url, headers, post_data, is_streaming=True
        )

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[False],
    ) -> Tuple[str, int, Any]: ...

    @overload
    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming: Literal[True],
    ) -> Tuple[HTTPResponse, int, Any]: ...

    def _request_internal(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data,
        is_streaming,
    ):
        if isinstance(post_data, str):
            post_data = post_data.encode("utf-8")

        req = self.urllibrequest.Request(
            url, post_data, cast(MutableMapping[str, str], headers)
        )

        if method not in ("get", "post"):
            req.get_method = lambda: method.upper()

        try:
            # use the custom proxy tied opener, if any.
            # otherwise, fall to the default urllib opener.
            response = (
                self._opener.open(req)
                if self._opener
                else self.urllibrequest.urlopen(req)
            )

            if is_streaming:
                rcontent = response
            else:
                rcontent = response.read()

            rcode = response.code
            headers = dict(response.info())
        except self.urlliberror.HTTPError as e:
            rcode = e.code
            rcontent = e.read()
            headers = dict(e.info())
        except (self.urlliberror.URLError, ValueError) as e:
            self._handle_request_error(e)
        lh = dict((k.lower(), v) for k, v in iter(dict(headers).items()))
        return rcontent, rcode, lh

    def _handle_request_error(self, e: Exception) -> NoReturn:
        msg = (
            "Unexpected error communicating with Stripe. "
            "If this problem persists, let us know at support@stripe.com."
        )
        msg = textwrap.fill(msg) + "\n\n(Network error: " + str(e) + ")"
        raise APIConnectionError(msg) from e

    def close(self):
        pass


class HTTPXClient(HTTPClient):
    name = "httpx"

    _client: Optional["HTTPXClientType"]

    def __init__(
        self,
        timeout: Optional[Union[float, "HTTPXTimeout"]] = 80,
        allow_sync_methods=False,
        _lib=None,  # used for internal unit testing
        **kwargs,
    ):
        super(HTTPXClient, self).__init__(**kwargs)

        if _lib is None:
            import httpx

            _lib = httpx
        self.httpx = _lib

        import anyio

        self.anyio = anyio

        kwargs = {}
        if self._verify_ssl_certs:
            kwargs["verify"] = ssl.create_default_context(
                cafile=stripe.ca_bundle_path
            )
        else:
            kwargs["verify"] = False

        self._client_async = self.httpx.AsyncClient(**kwargs)
        self._client = None
        if allow_sync_methods:
            self._client = self.httpx.Client(**kwargs)
        self._timeout = timeout

    def sleep_async(self, secs):
        return self.anyio.sleep(secs)

    def _get_request_args_kwargs(
        self, method: str, url: str, headers: Mapping[str, str], post_data
    ):
        kwargs = {}

        if self._proxy:
            kwargs["proxies"] = self._proxy

        if self._timeout:
            kwargs["timeout"] = self._timeout
        return [
            (method, url),
            {"headers": headers, "data": post_data or {}, **kwargs},
        ]

    def request(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data=None,
    ) -> Tuple[bytes, int, Mapping[str, str]]:
        if self._client is None:
            raise RuntimeError(
                "Stripe: HTTPXClient was initialized with allow_sync_methods=False, "
                "so it cannot be used for synchronous requests."
            )
        args, kwargs = self._get_request_args_kwargs(
            method, url, headers, post_data
        )
        try:
            response = self._client.request(*args, **kwargs)
        except Exception as e:
            self._handle_request_error(e)

        content = response.content
        status_code = response.status_code
        response_headers = response.headers
        return content, status_code, response_headers

    async def request_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data=None,
    ) -> Tuple[bytes, int, Mapping[str, str]]:
        args, kwargs = self._get_request_args_kwargs(
            method, url, headers, post_data
        )
        try:
            response = await self._client_async.request(*args, **kwargs)
        except Exception as e:
            self._handle_request_error(e)

        content = response.content
        status_code = response.status_code
        response_headers = response.headers
        return content, status_code, response_headers

    def _handle_request_error(self, e: Exception) -> NoReturn:
        msg = (
            "Unexpected error communicating with Stripe. If this "
            "problem persists, let us know at support@stripe.com."
        )
        err = "A %s was raised" % (type(e).__name__,)
        should_retry = True

        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise APIConnectionError(msg, should_retry=should_retry) from e

    def request_stream(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[Iterable[bytes], int, Mapping[str, str]]:
        if self._client is None:
            raise RuntimeError(
                "Stripe: HTTPXClient was not initialized with allow_sync_methods=True, "
                "so it cannot be used for synchronous requests."
            )
        args, kwargs = self._get_request_args_kwargs(
            method, url, headers, post_data
        )
        try:
            response = self._client.send(
                request=self._client_async.build_request(*args, **kwargs),
                stream=True,
            )
        except Exception as e:
            self._handle_request_error(e)
        content = response.iter_bytes()
        status_code = response.status_code
        headers = response.headers

        return content, status_code, headers

    async def request_stream_async(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[AsyncIterable[bytes], int, Mapping[str, str]]:
        args, kwargs = self._get_request_args_kwargs(
            method, url, headers, post_data
        )
        try:
            response = await self._client_async.send(
                request=self._client_async.build_request(*args, **kwargs),
                stream=True,
            )
        except Exception as e:
            self._handle_request_error(e)
        content = response.aiter_bytes()
        status_code = response.status_code
        headers = response.headers

        return content, status_code, headers

    def close(self):
        if self._client is not None:
            self._client.close()

    async def close_async(self):
        await self._client_async.aclose()


class AIOHTTPClient(HTTPClient):
    name = "aiohttp"

    def __init__(
        self,
        timeout: Optional[Union[float, "AIOHTTPTimeout"]] = 80,
        _lib=None,  # used for internal unit testing
        session=None,
        connector=None,
        **kwargs,
    ):
        super(AIOHTTPClient, self).__init__(**kwargs)

        if _lib is None:
            import aiohttp

            _lib = aiohttp

        self.aiohttp = _lib

        self._timeout = timeout
        self._user_session = session
        self._user_connector = connector
        self._internally_managed_session = session is None
        self._cached_session = None

    @property
    def _session(self):
        if self._cached_session is None:
            if self._user_session:
                self._cached_session = self._user_session
            else:
                kwargs = {}
                if self._user_connector:
                    kwargs["connector"] = self._user_connector
                elif self._verify_ssl_certs:
                    ssl_context = ssl.create_default_context(
                        cafile=stripe.ca_bundle_path
                    )
                    kwargs["connector"] = self.aiohttp.TCPConnector(
                        ssl=ssl_context
                    )
                else:
                    kwargs["connector"] = self.aiohttp.TCPConnector(
                        verify_ssl=False
                    )

                self._cached_session = self.aiohttp.ClientSession(**kwargs)

        return self._cached_session

    def sleep_async(self, secs):
        return asyncio.sleep(secs)

    def request(self) -> Tuple[bytes, int, Mapping[str, str]]:
        raise NotImplementedError(
            "AIOHTTPClient does not support synchronous requests."
        )

    def _get_request_args_kwargs(
        self, method: str, url: str, headers: Mapping[str, str], post_data
    ):
        args = (method, url)
        kwargs = {}
        if self._proxy:
            if self._proxy["http"] != self._proxy["https"]:
                raise ValueError(
                    "AIOHTTPClient does not support different proxies for HTTP and HTTPS."
                )
            kwargs["proxy"] = self._proxy["https"]
        if self._timeout:
            kwargs["timeout"] = self._timeout

        kwargs["headers"] = headers
        kwargs["data"] = post_data
        return args, kwargs

    async def request_async(
        self,
        method: str,
        url: str,
        headers: Mapping[str, str],
        post_data=None,
    ) -> Tuple[bytes, int, Mapping[str, str]]:
        (
            content,
            status_code,
            response_headers,
        ) = await self.request_stream_async(
            method, url, headers, post_data=post_data
        )

        return (await content.read()), status_code, response_headers

    def _handle_request_error(self, e: Exception) -> NoReturn:
        msg = (
            "Unexpected error communicating with Stripe. If this "
            "problem persists, let us know at support@stripe.com."
        )
        err = "A %s was raised" % (type(e).__name__,)
        should_retry = True

        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise APIConnectionError(msg, should_retry=should_retry) from e

    def request_stream(self) -> Tuple[Iterable[bytes], int, Mapping[str, str]]:
        raise NotImplementedError(
            "AIOHTTPClient does not support synchronous requests."
        )

    async def request_stream_async(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple["AIOHTTPStreamReader", int, Mapping[str, str]]:
        args, kwargs = self._get_request_args_kwargs(
            method, url, headers, post_data
        )
        try:
            response = await self._session.request(*args, **kwargs)
        except Exception as e:
            self._handle_request_error(e)

        content = response.content
        status_code = response.status
        response_headers = response.headers
        return content, status_code, response_headers

    def close(self):
        pass

    async def close_async(self):
        if self._internally_managed_session:
            await self._session.close()


class NoImportFoundAsyncClient(HTTPClient):
    def __init__(self, **kwargs):
        super(NoImportFoundAsyncClient, self).__init__(**kwargs)

    @staticmethod
    def raise_async_client_import_error() -> Never:
        raise ImportError(
            (
                "Import httpx not found. To make async http requests,"
                "You must either install httpx or define your own"
                "async http client by subclassing stripe.HTTPClient"
                "and setting stripe.default_http_client to an instance of it."
            )
        )

    async def request_async(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ) -> Tuple[bytes, int, Mapping[str, str]]:
        self.raise_async_client_import_error()

    async def request_stream_async(
        self, method: str, url: str, headers: Mapping[str, str], post_data=None
    ):
        self.raise_async_client_import_error()

    async def close_async(self):
        self.raise_async_client_import_error()


# --- Client resolution ---
# Detect available HTTP libraries at module load time so the expensive imports
# (e.g. requests, httpx) happen during Python's init phase rather than when
# StripeClient() is constructed. This matters in environments like AWS Lambda
# where module loading has a generous timeout (10s) but handler invocation
# does not (often 3s).
#
# Sync client precedence:
#   1. Urlfetch (Google App Engine — if present, you probably want it)
#   2. Requests (popular, top priority outside GAE)
#   3. Pycurl (verifies SSL certs, but less preferred than Requests)
#   4. urllib (stdlib fallback, basically always present)
#
# Async client precedence:
#   1. httpx + anyio (both required)
#   2. aiohttp
#   3. NoImportFoundAsyncClient (raises on use)
#
# To add a new client: define the class above, then add it to the appropriate
# cascade below. The resolved class is stored directly — new_default_http_client()
# and new_http_client_async_fallback() just call it.


def _resolve_sync_client():
    try:
        from google.appengine.api import urlfetch  # type: ignore # noqa: F401

        return UrlFetchClient
    except ImportError:
        pass

    try:
        import requests  # noqa: F401

        return RequestsClient
    except ImportError:
        pass

    try:
        import pycurl  # type: ignore # noqa: F401

        return PycurlClient
    except ImportError:
        pass

    return UrllibClient


def _resolve_async_client():
    try:
        import httpx  # noqa: F401
        import anyio  # noqa: F401

        return HTTPXClient
    except ImportError:
        pass

    try:
        import aiohttp  # noqa: F401

        return AIOHTTPClient
    except ImportError:
        pass

    return NoImportFoundAsyncClient


# Called at module load time so HTTP library imports happen during Python's init
# phase rather than inside StripeClient.__init__().
_default_sync_client = _resolve_sync_client()
_default_async_client = _resolve_async_client()



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from typing import List
from typing_extensions import Literal, NotRequired, TypedDict


class CustomerEvaluationCreateParams(RequestOptions):
    evaluation_context: List["CustomerEvaluationCreateParamsEvaluationContext"]
    """
    Array of context entries for the evaluation.
    """
    event_type: Literal["login", "registration"]
    """
    The type of evaluation event.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """


class CustomerEvaluationCreateParamsEvaluationContext(TypedDict):
    client_details: NotRequired[
        "CustomerEvaluationCreateParamsEvaluationContextClientDetails"
    ]
    """
    Client details context.
    """
    customer_details: NotRequired[
        "CustomerEvaluationCreateParamsEvaluationContextCustomerDetails"
    ]
    """
    Customer details context.
    """
    type: Literal["client_details", "customer_details"]
    """
    The type of context entry.
    """


class CustomerEvaluationCreateParamsEvaluationContextClientDetails(TypedDict):
    data: NotRequired[
        "CustomerEvaluationCreateParamsEvaluationContextClientDetailsData"
    ]
    """
    Raw client metadata fallback in case a Radar Session is unavailable.
    """
    radar_session: NotRequired[str]
    """
    ID for the Radar Session. Required unless data is provided.
    """


class CustomerEvaluationCreateParamsEvaluationContextClientDetailsData(
    TypedDict,
):
    ip: str
    """
    The end user's IP address. Used for proxy detection and IP-clustering signals.
    """
    referrer: NotRequired[str]
    """
    The referring URL of the login or registration page.
    """
    user_agent: NotRequired[str]
    """
    The User-Agent HTTP header.
    """


class CustomerEvaluationCreateParamsEvaluationContextCustomerDetails(
    TypedDict
):
    customer: NotRequired[str]
    """
    The ID of an existing Customer.
    """
    customer_data: NotRequired[
        "CustomerEvaluationCreateParamsEvaluationContextCustomerDetailsCustomerData"
    ]
    """
    Customer data
    """


class CustomerEvaluationCreateParamsEvaluationContextCustomerDetailsCustomerData(
    TypedDict,
):
    email: NotRequired[str]
    """
    Customer email
    """
    name: NotRequired[str]
    """
    Customer name
    """
    phone: NotRequired[str]
    """
    Customer phone
    """



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from typing import List
from typing_extensions import Literal, NotRequired, TypedDict


class MeterCreateParams(RequestOptions):
    customer_mapping: NotRequired["MeterCreateParamsCustomerMapping"]
    """
    Fields that specify how to map a meter event to a customer.
    """
    default_aggregation: "MeterCreateParamsDefaultAggregation"
    """
    The default settings to aggregate a meter's events with.
    """
    dimension_payload_keys: NotRequired[List[str]]
    """
    Set of keys that will be used to group meter events by. Each key must be present in the event payload.
    """
    display_name: str
    """
    The meter's name. Not visible to the customer.
    """
    event_name: str
    """
    The name of the meter event to record usage for. Corresponds with the `event_name` field on meter events.
    """
    event_time_window: NotRequired[Literal["day", "hour"]]
    """
    The time window which meter events have been pre-aggregated for, if any.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    value_settings: NotRequired["MeterCreateParamsValueSettings"]
    """
    Fields that specify how to calculate a meter event's value.
    """


class MeterCreateParamsCustomerMapping(TypedDict):
    event_payload_key: str
    """
    The key in the meter event payload to use for mapping the event to a customer.
    """
    type: Literal["by_id"]
    """
    The method for mapping a meter event to a customer. Must be `by_id`.
    """


class MeterCreateParamsDefaultAggregation(TypedDict):
    formula: Literal["count", "last", "sum"]
    """
    Specifies how events are aggregated.
    """


class MeterCreateParamsValueSettings(TypedDict):
    event_payload_key: str
    """
    The key in the usage event payload to use as the value for this meter. For example, if the event payload contains usage on a `bytes_used` field, then set the event_payload_key to "bytes_used".
    """



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from stripe._stripe_object import UntypedStripeObject
from typing import Dict, List
from typing_extensions import NotRequired, TypedDict


class ReaderRefundPaymentParams(RequestOptions):
    amount: NotRequired[int]
    """
    A positive integer in __cents__ representing how much of this charge to refund.
    """
    charge: NotRequired[str]
    """
    ID of the Charge to refund.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    metadata: NotRequired["Dict[str, str]|UntypedStripeObject[str]"]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format. Individual keys can be unset by posting an empty value to them. All keys can be unset by posting an empty value to `metadata`.
    """
    payment_intent: NotRequired[str]
    """
    ID of the PaymentIntent to refund.
    """
    refund_application_fee: NotRequired[bool]
    """
    Boolean indicating whether the application fee should be refunded when refunding this charge. If a full charge refund is given, the full application fee will be refunded. Otherwise, the application fee will be refunded in an amount proportional to the amount of the charge refunded. An application fee can be refunded only by the application that created the charge.
    """
    refund_payment_config: NotRequired[
        "ReaderRefundPaymentParamsRefundPaymentConfig"
    ]
    """
    Configuration overrides for this refund, such as customer cancellation settings.
    """
    reverse_transfer: NotRequired[bool]
    """
    Boolean indicating whether the transfer should be reversed when refunding this charge. The transfer will be reversed proportionally to the amount being refunded (either the entire or partial amount). A transfer can be reversed only by the application that created the charge.
    """


class ReaderRefundPaymentParamsRefundPaymentConfig(TypedDict):
    enable_customer_cancellation: NotRequired[bool]
    """
    Enables cancel button on transaction screens.
    """



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._stripe_service import StripeService
from stripe._util import sanitize_id
from typing import Optional, cast
from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:
    from stripe._list_object import ListObject
    from stripe._request_options import RequestOptions
    from stripe.financial_connections._institution import Institution
    from stripe.params.financial_connections._institution_list_params import (
        InstitutionListParams,
    )
    from stripe.params.financial_connections._institution_retrieve_params import (
        InstitutionRetrieveParams,
    )


class InstitutionService(StripeService):
    def list(
        self,
        params: Optional["InstitutionListParams"] = None,
        options: Optional["RequestOptions"] = None,
    ) -> "ListObject[Institution]":
        """
        Returns a list of Financial Connections Institution objects.
        """
        return cast(
            "ListObject[Institution]",
            self._request(
                "get",
                "/v1/financial_connections/institutions",
                base_address="api",
                params=params,
                options=options,
            ),
        )

    async def list_async(
        self,
        params: Optional["InstitutionListParams"] = None,
        options: Optional["RequestOptions"] = None,
    ) -> "ListObject[Institution]":
        """
        Returns a list of Financial Connections Institution objects.
        """
        return cast(
            "ListObject[Institution]",
            await self._request_async(
                "get",
                "/v1/financial_connections/institutions",
                base_address="api",
                params=params,
                options=options,
            ),
        )

    def retrieve(
        self,
        institution: str,
        params: Optional["InstitutionRetrieveParams"] = None,
        options: Optional["RequestOptions"] = None,
    ) -> "Institution":
        """
        Retrieves the details of a Financial Connections Institution.
        """
        return cast(
            "Institution",
            self._request(
                "get",
                "/v1/financial_connections/institutions/{institution}".format(
                    institution=sanitize_id(institution),
                ),
                base_address="api",
                params=params,
                options=options,
            ),
        )

    async def retrieve_async(
        self,
        institution: str,
        params: Optional["InstitutionRetrieveParams"] = None,
        options: Optional["RequestOptions"] = None,
    ) -> "Institution":
        """
        Retrieves the details of a Financial Connections Institution.
        """
        return cast(
            "Institution",
            await self._request_async(
                "get",
                "/v1/financial_connections/institutions/{institution}".format(
                    institution=sanitize_id(institution),
                ),
                base_address="api",
                params=params,
                options=options,
            ),
        )



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from typing import List
from typing_extensions import Literal, NotRequired, TypedDict


class FileListParams(RequestOptions):
    created: NotRequired["FileListParamsCreated|int"]
    """
    Only return files that were created during the given date interval.
    """
    ending_before: NotRequired[str]
    """
    A cursor for use in pagination. `ending_before` is an object ID that defines your place in the list. For instance, if you make a list request and receive 100 objects, starting with `obj_bar`, your subsequent call can include `ending_before=obj_bar` in order to fetch the previous page of the list.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    limit: NotRequired[int]
    """
    A limit on the number of objects to be returned. Limit can range between 1 and 100, and the default is 10.
    """
    purpose: NotRequired[
        Literal[
            "account_requirement",
            "additional_verification",
            "business_icon",
            "business_logo",
            "customer_signature",
            "dispute_evidence",
            "document_provider_identity_document",
            "finance_report_run",
            "financial_account_statement",
            "identity_document",
            "identity_document_downloadable",
            "issuing_regulatory_reporting",
            "pci_document",
            "platform_terms_of_service",
            "selfie",
            "sigma_scheduled_query",
            "tax_document_user_upload",
            "terminal_android_apk",
            "terminal_reader_splashscreen",
            "terminal_wifi_certificate",
            "terminal_wifi_private_key",
        ]
    ]
    """
    Filter queries by the file purpose. If you don't provide a purpose, the queries return unfiltered files.
    """
    starting_after: NotRequired[str]
    """
    A cursor for use in pagination. `starting_after` is an object ID that defines your place in the list. For instance, if you make a list request and receive 100 objects, ending with `obj_foo`, your subsequent call can include `starting_after=obj_foo` in order to fetch the next page of the list.
    """


class FileListParamsCreated(TypedDict):
    gt: NotRequired[int]
    """
    Minimum value to filter by (exclusive)
    """
    gte: NotRequired[int]
    """
    Minimum value to filter by (inclusive)
    """
    lt: NotRequired[int]
    """
    Maximum value to filter by (exclusive)
    """
    lte: NotRequired[int]
    """
    Maximum value to filter by (inclusive)
    """



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._stripe_object import StripeObject
from typing import ClassVar, List, Optional
from typing_extensions import Literal


class AccountSignals(StripeObject):
    """
    The Account Signals API provides risk related signals that can be used to better manage risks.
    """

    OBJECT_NAME: ClassVar[Literal["account_signals"]] = "account_signals"

    class Delinquency(StripeObject):
        class Indicator(StripeObject):
            description: str
            """
            A brief explanation of how this indicator contributed to the delinquency probability.
            """
            impact: Literal[
                "decrease", "neutral", "slight_increase", "strong_increase"
            ]
            """
            The effect this indicator had on the overall risk level.
            """
            indicator: Literal[
                "account_balance",
                "aov",
                "charge_concentration",
                "dispute_window",
                "disputes",
                "duplicates",
                "exposure",
                "firmographic",
                "lifetime_metrics",
                "payment_processing",
                "payment_volume",
                "payouts",
                "refunds",
                "related_accounts",
                "tenure",
                "transfers",
            ]
            """
            The name of the specific indicator used in the risk assessment.
            """

        evaluated_at: Optional[int]
        """
        Time at which the signal was evaluated, measured in seconds since the Unix epoch.
        """
        indicators: Optional[List[Indicator]]
        """
        Array of objects representing individual factors that contributed to the calculated probability of delinquency.
        """
        probability: Optional[float]
        """
        The probability of delinquency. Can be between 0.00 and 100.00
        """
        risk_level: Literal[
            "elevated", "highest", "low", "normal", "not_assessed", "unknown"
        ]
        """
        Categorical assessment of the delinquency risk based on probability.
        """
        signal_id: Optional[str]
        """
        Unique identifier for the delinquency signal.
        """
        _inner_class_types = {"indicators": Indicator}

    class FraudIntent(StripeObject):
        class Indicator(StripeObject):
            description: str
            """
            A brief explanation of how this indicator contributed to the delinquency probability.
            """
            impact: Literal[
                "decrease", "neutral", "slight_increase", "strong_increase"
            ]
            """
            The effect this indicator had on the overall risk level.
            """
            indicator: Literal[
                "bank_account",
                "business_information_and_account_activity",
                "disputes",
                "failures",
                "geo_location",
                "other",
                "other_related_accounts",
                "other_transaction_activity",
                "owner_email",
                "web_presence",
            ]
            """
            The name of the specific indicator used in the risk assessment.
            """

        evaluated_at: Optional[int]
        """
        Time at which the signal was evaluated, measured in seconds since the Unix epoch.
        """
        indicators: Optional[List[Indicator]]
        """
        Array of objects representing individual factors that contributed to the calculated probability of fraud intent.
        """
        probability: Optional[float]
        """
        The probability of fraud intent. Can be between 0.00 and 100.00
        """
        risk_level: Literal[
            "elevated", "highest", "low", "normal", "not_assessed", "unknown"
        ]
        """
        Categorical assessment of the fraud intent risk based on probability.
        """
        signal_id: Optional[str]
        """
        Unique identifier for the fraud intent signal.
        """
        _inner_class_types = {"indicators": Indicator}

    account: str
    """
    The account for which the signals belong to.
    """
    delinquency: Optional[Delinquency]
    """
    The delinquency signal of the account.
    """
    fraud_intent: Optional[FraudIntent]
    """
    The fraud intent signal of the account.
    """
    livemode: bool
    """
    Has the value `true` if the object exists in live mode or the value `false` if the object exists in test mode.
    """
    object: Literal["account_signals"]
    """
    String representing the object's type. Objects of the same type share the same value.
    """
    _inner_class_types = {
        "delinquency": Delinquency,
        "fraud_intent": FraudIntent,
    }



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._api_resource import APIResource
from stripe._stripe_object import StripeObject
from typing import ClassVar, List, Optional
from typing_extensions import Literal, Unpack, TYPE_CHECKING

if TYPE_CHECKING:
    from stripe.delegated_checkout._order_event import OrderEvent
    from stripe.params.delegated_checkout._order_retrieve_params import (
        OrderRetrieveParams,
    )


class Order(APIResource["Order"]):
    """
    An order represents the post-checkout lifecycle of a delegated checkout purchase.
    """

    OBJECT_NAME: ClassVar[Literal["delegated_checkout.order"]] = (
        "delegated_checkout.order"
    )

    class LineItem(StripeObject):
        class ProductDetails(StripeObject):
            description: Optional[str]
            """
            The item description.
            """
            images: Optional[List[str]]
            """
            The item images.
            """
            title: str
            """
            The item title.
            """

        class Quantity(StripeObject):
            current: int
            """
            The current quantity.
            """
            ordered: int
            """
            The ordered quantity.
            """
            shipped: int
            """
            The shipped quantity.
            """

        class Totals(StripeObject):
            base_amount: Optional[int]
            """
            The base amount for the line item.
            """
            discount: Optional[int]
            """
            The discount amount for the line item.
            """
            subtotal: Optional[int]
            """
            The subtotal amount for the line item.
            """
            tax: Optional[int]
            """
            The tax amount for the line item.
            """
            total: Optional[int]
            """
            The total amount for the line item.
            """

        key: str
        """
        The order line item key.
        """
        product_details: ProductDetails
        quantity: Quantity
        sku_id: str
        """
        The SKU ID of the line item.
        """
        totals: Optional[Totals]
        """
        The totals for this line item.
        """
        unit_amount: int
        """
        The line item unit amount.
        """
        _inner_class_types = {
            "product_details": ProductDetails,
            "quantity": Quantity,
            "totals": Totals,
        }

    class Totals(StripeObject):
        discount: Optional[int]
        """
        The discount amount for the order.
        """
        fulfillment: Optional[int]
        """
        The fulfillment amount for the order.
        """
        subtotal: Optional[int]
        """
        The subtotal amount for the order.
        """
        tax: Optional[int]
        """
        The tax amount for the order.
        """
        total: Optional[int]
        """
        The total amount for the order.
        """

    created: int
    """
    Time at which the object was created. Measured in seconds since the Unix epoch.
    """
    currency: str
    """
    Three-letter [ISO currency code](https://www.iso.org/iso-4217-currency-codes.html), in lowercase. Must be a [supported currency](https://stripe.com/docs/currencies).
    """
    id: str
    """
    Unique identifier for the object.
    """
    latest_order_event: Optional["OrderEvent"]
    """
    The latest order event for this order.
    """
    line_items: List[LineItem]
    """
    The line items in this order.
    """
    livemode: bool
    """
    If the object exists in live mode, the value is `true`. If the object exists in test mode, the value is `false`.
    """
    object: Literal["delegated_checkout.order"]
    """
    String representing the object's type. Objects of the same type share the same value.
    """
    permalink_url: Optional[str]
    """
    The permalink URL for this order.
    """
    requested_session: str
    """
    The requested session associated with this order.
    """
    seller_reference: Optional[str]
    """
    The seller reference for this order.
    """
    totals: Optional[Totals]
    """
    The totals for this order.
    """

    @classmethod
    def retrieve(
        cls, id: str, **params: Unpack["OrderRetrieveParams"]
    ) -> "Order":
        """
        Retrieves a delegated checkout order.
        """
        instance = cls(id, **params)
        instance.refresh()
        return instance

    @classmethod
    async def retrieve_async(
        cls, id: str, **params: Unpack["OrderRetrieveParams"]
    ) -> "Order":
        """
        Retrieves a delegated checkout order.
        """
        instance = cls(id, **params)
        await instance.refresh_async()
        return instance

    _inner_class_types = {"line_items": LineItem, "totals": Totals}



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from decimal import Decimal
from stripe._stripe_object import StripeObject
from stripe.v2._amount import Amount
from typing import ClassVar, List, Optional
from typing_extensions import Literal


class Alert(StripeObject):
    """
    A health alert represents a detected problem affecting a merchant's Stripe integration.
    """

    OBJECT_NAME: ClassVar[Literal["v2.core.health.alert"]] = (
        "v2.core.health.alert"
    )

    class ApiError(StripeObject):
        class TopImpactedAccount(StripeObject):
            account: str
            """
            The account ID of the impacted connected account.
            """
            impacted_requests: int
            """
            The number of impacted requests.
            """
            impacted_requests_percentage: Optional[Decimal]
            """
            The percentage of impacted requests.
            """
            _field_encodings = {
                "impacted_requests_percentage": "decimal_string",
            }

        canonical_path: str
        """
        The canonical path.
        """
        error_code: Optional[str]
        """
        The error code.
        """
        http_method: Literal["DELETE", "GET", "POST", "PUT"]
        """
        The HTTP method.
        """
        http_status: str
        """
        The HTTP status.
        """
        impacted_requests: int
        """
        The number of impacted requests.
        """
        impacted_requests_percentage: Optional[Decimal]
        """
        The percentage of impacted requests.
        """
        top_impacted_accounts: Optional[List[TopImpactedAccount]]
        """
        The top impacted connected accounts (only for platforms).
        """
        _inner_class_types = {"top_impacted_accounts": TopImpactedAccount}
        _field_encodings = {"impacted_requests_percentage": "decimal_string"}

    class ApiLatency(StripeObject):
        class TopImpactedAccount(StripeObject):
            account: str
            """
            The account ID of the impacted connected account.
            """
            impacted_requests: int
            """
            The number of impacted requests.
            """
            impacted_requests_percentage: Optional[Decimal]
            """
            The percentage of impacted requests.
            """
            _field_encodings = {
                "impacted_requests_percentage": "decimal_string",
            }

        canonical_path: str
        """
        The canonical path.
        """
        http_method: Literal["DELETE", "GET", "POST", "PUT"]
        """
        The HTTP method.
        """
        http_status: str
        """
        The HTTP status.
        """
        impacted_requests: int
        """
        The number of impacted requests.
        """
        impacted_requests_percentage: Optional[Decimal]
        """
        The percentage of impacted requests.
        """
        top_impacted_accounts: Optional[List[TopImpactedAccount]]
        """
        The top impacted connected accounts (only for platforms).
        """
        _inner_class_types = {"top_impacted_accounts": TopImpactedAccount}
        _field_encodings = {"impacted_requests_percentage": "decimal_string"}

    class AuthorizationRateDrop(StripeObject):
        class Dimension(StripeObject):
            issuer: Optional[str]
            """
            Populated when type is issuer.
            """
            type: Literal["issuer"]
            """
            The type of the dimension. Determines which field in dimension_details is populated.
            """

        charge_type: Literal["money_moving", "validation"]
        """
        The type of the charge.
        """
        current_percentage: Decimal
        """
        The current authorization rate percentage.
        """
        dimensions: Optional[List[Dimension]]
        """
        Dimensions that describe what subset of payments are impacted.
        """
        payment_method_type: Literal[
            "acss_debit",
            "affirm",
            "afterpay_clearpay",
            "alipay",
            "alma",
            "amazon_pay",
            "apple_pay",
            "au_becs_debit",
            "bacs_debit",
            "bancontact",
            "billie",
            "blik",
            "boleto",
            "card",
            "card_present",
            "cartes_bancaires",
            "cashapp",
            "crypto",
            "dummy_passthrough_card",
            "eps",
            "fpx",
            "giropay",
            "grabpay",
            "ideal",
            "interac_present",
            "kakao_pay",
            "klarna",
            "konbini",
            "kriya",
            "kr_card",
            "link",
            "mb_way",
            "mobilepay",
            "mondu",
            "multibanco",
            "naver_pay",
            "ng_bank",
            "ng_bank_transfer",
            "ng_card",
            "ng_market",
            "ng_ussd",
            "ng_wallet",
            "oxxo",
            "p24",
            "paper_check",
            "payco",
            "paynow",
            "paypal",
            "paypay",
            "payto",
            "pay_by_bank",
            "pix",
            "promptpay",
            "rechnung",
            "revolut_pay",
            "samsung_pay",
            "satispay",
            "scalapay",
            "sepa_debit",
            "sequra",
            "sofort",
            "sunbit",
            "swish",
            "twint",
            "upi",
            "us_bank_account",
            "vipps",
            "wechat_pay",
            "zip",
        ]
        """
        The type of the payment method.
        """
        previous_percentage: Decimal
        """
        The previous authorization rate percentage.
        """
        _inner_class_types = {"dimensions": Dimension}
        _field_encodings = {
            "current_percentage": "decimal_string",
            "previous_percentage": "decimal_string",
        }

    class DocumentationLink(StripeObject):
        label: str
        """
        A human-readable label for the link.
        """
        url: str
        """
        The URL of the documentation.
        """

    class ElementsError(StripeObject):
        element_type: Optional[Literal["expressCheckout", "payment"]]
        """
        The type of the element.
        """
        impacted_sessions: int
        """
        The number of impacted sessions.
        """

    class EventGenerationFailure(StripeObject):
        class RelatedObject(StripeObject):
            id: str
            """
            The ID of the related object (e.g., "pi_...").
            """
            type: str
            """
            The type of the related object (e.g., "payment_intent").
            """
            url: str
            """
            The API URL for the related object (e.g., "/v1/payment_intents/pi_...").
            """

        context: Optional[str]
        """
        The context the event should have been generated for. Only present when the account is a connected account.
        """
        event_type: str
        """
        The type of event that Stripe failed to generate.
        """
        related_object: RelatedObject
        """
        The related object details.
        """
        _inner_class_types = {"related_object": RelatedObject}

    class FraudRate(StripeObject):
        attack_type: Literal["spike", "sustained_attack"]
        """
        Fraud attack type.
        """
        impacted_requests: int
        """
        The number of impacted requests which are detected.
        """
        realized_fraud_amount: Amount
        """
        Estimated aggregated amount for the impacted requests.
        """

    class InvoiceCountDropped(StripeObject):
        observed_count: Decimal
        """
        The observed number of invoices within the time window.
        """
        threshold_count: Decimal
        """
        The expected threshold number of invoices within the time window.
        """
        time_window: str
        """
        The size of the observation time window.
        """
        _field_encodings = {
            "observed_count": "decimal_string",
            "threshold_count": "decimal_string",
        }

    class IssuingAuthorizationRequestErrors(StripeObject):
        approved_amount: Optional[Amount]
        """
        Estimated aggregated amount for the approved requests.
        """
        approved_impacted_requests: Optional[int]
        """
        The number of approved requests which are impacted.
        """
        declined_amount: Optional[Amount]
        """
        Estimated aggregated amount for the declined requests.
        """
        declined_impacted_requests: Optional[int]
        """
        The number of declined requests which are impacted.
        """

    class IssuingAuthorizationRequestTimeout(StripeObject):
        approved_amount: Optional[Amount]
        """
        Estimated aggregated amount for the approved requests.
        """
        approved_impacted_requests: Optional[int]
        """
        The number of approved requests which are impacted.
        """
        declined_amount: Optional[Amount]
        """
        Estimated aggregated amount for the declined requests.
        """
        declined_impacted_requests: Optional[int]
        """
        The number of declined requests which are impacted.
        """

    class MeterEventSummariesDelayed(StripeObject):
        ingestion_method: Optional[Literal["import_sets"]]
        """
        The ingestion method.
        """

    class MetronomeNotificationLatency(StripeObject):
        pipeline: Literal[
            "configuration_triggered",
            "high_cardinality_usage_triggered",
            "standard_usage_triggered",
            "time_triggered",
        ]
        """
        The impacted Metronome billing pipeline.
        """

    class PaymentMethodError(StripeObject):
        class TopImpactedAccount(StripeObject):
            account: str
            """
            The account ID of the impacted connected account.
            """
            impacted_requests: int
            """
            The number of impacted requests.
            """
            impacted_requests_percentage: Optional[Decimal]
            """
            The percentage of impacted requests.
            """
            _field_encodings = {
                "impacted_requests_percentage": "decimal_string",
            }

        error_code: Optional[str]
        """
        The returned error code.
        """
        impacted_requests: int
        """
        The number of impacted requests.
        """
        impacted_requests_percentage: Optional[Decimal]
        """
        The percentage of impacted requests.
        """
        payment_method_type: Literal[
            "acss_debit",
            "affirm",
            "afterpay_clearpay",
            "alipay",
            "alma",
            "amazon_pay",
            "apple_pay",
            "au_becs_debit",
            "bacs_debit",
            "bancontact",
            "billie",
            "blik",
            "boleto",
            "card",
            "card_present",
            "cartes_bancaires",
            "cashapp",
            "crypto",
            "dummy_passthrough_card",
            "eps",
            "fpx",
            "giropay",
            "grabpay",
            "ideal",
            "interac_present",
            "kakao_pay",
            "klarna",
            "konbini",
            "kriya",
            "kr_card",
            "link",
            "mb_way",
            "mobilepay",
            "mondu",
            "multibanco",
            "naver_pay",
            "ng_bank",
            "ng_bank_transfer",
            "ng_card",
            "ng_market",
            "ng_ussd",
            "ng_wallet",
            "oxxo",
            "p24",
            "paper_check",
            "payco",
            "paynow",
            "paypal",
            "paypay",
            "payto",
            "pay_by_bank",
            "pix",
            "promptpay",
            "rechnung",
            "revolut_pay",
            "samsung_pay",
            "satispay",
            "scalapay",
            "sepa_debit",
            "sequra",
            "sofort",
            "sunbit",
            "swish",
            "twint",
            "upi",
            "us_bank_account",
            "vipps",
            "wechat_pay",
            "zip",
        ]
        """
        The type of the payment method.
        """
        top_impacted_accounts: Optional[List[TopImpactedAccount]]
        """
        The top impacted connected accounts (only for platforms).
        """
        _inner_class_types = {"top_impacted_accounts": TopImpactedAccount}
        _field_encodings = {"impacted_requests_percentage": "decimal_string"}

    class SepaDebitDelayed(StripeObject):
        impacted_payments: int
        """
        The number of impacted payments.
        """
        impacted_payments_percentage: Decimal
        """
        The percentage of impacted payments.
        """
        _field_encodings = {"impacted_payments_percentage": "decimal_string"}

    class TrafficVolumeDrop(StripeObject):
        actual_traffic: int
        """
        The total volume of payment requests within the latest observation time window.
        """
        canonical_path: Optional[str]
        """
        The canonical path.
        """
        expected_traffic: Optional[int]
        """
        The expected volume of payment requests within the latest observation time window.
        """
        time_window: str
        """
        The size of the observation time window.
        """

    class WebhookLatency(StripeObject):
        impacted_requests: int
        """
        The number of impacted requests.
        """

    api_error: Optional[ApiError]
    """
    Populated when type is api_error.
    """
    api_latency: Optional[ApiLatency]
    """
    Populated when type is api_latency.
    """
    authorization_rate_drop: Optional[AuthorizationRateDrop]
    """
    Populated when type is authorization_rate_drop.
    """
    created: str
    """
    Time at which the health alert was created.
    """
    documentation_links: List[DocumentationLink]
    """
    Links to relevant documentation for diagnosing and resolving the alert.
    """
    elements_error: Optional[ElementsError]
    """
    Populated when type is elements_error.
    """
    event_generation_failure: Optional[EventGenerationFailure]
    """
    Populated when type is event_generation_failure.
    """
    fraud_rate: Optional[FraudRate]
    """
    Populated when type is fraud_rate.
    """
    grouping_key: str
    """
    The grouping key for the alert.
    """
    grouping_type: Literal["incident", "standalone"]
    """
    Whether the alert is linked to an incident or is a self-contained problem.
    """
    id: str
    """
    Unique identifier for the health alert.
    """
    invoice_count_dropped: Optional[InvoiceCountDropped]
    """
    Populated when type is invoice_count_dropped.
    """
    issuing_authorization_request_errors: Optional[
        IssuingAuthorizationRequestErrors
    ]
    """
    Populated when type is issuing_authorization_request_errors.
    """
    issuing_authorization_request_timeout: Optional[
        IssuingAuthorizationRequestTimeout
    ]
    """
    Populated when type is issuing_authorization_request_timeout.
    """
    livemode: bool
    """
    Has the value `true` if the object exists in live mode or the value `false` if the object exists in test mode.
    """
    meter_event_summaries_delayed: Optional[MeterEventSummariesDelayed]
    """
    Populated when type is meter_event_summaries_delayed.
    """
    metronome_notification_latency: Optional[MetronomeNotificationLatency]
    """
    Populated when type is metronome_notification_latency.
    """
    object: Literal["v2.core.health.alert"]
    """
    String representing the object's type. Objects of the same type share the same value of the object field.
    """
    payment_method_error: Optional[PaymentMethodError]
    """
    Populated when type is payment_method_error.
    """
    resolved_at: Optional[str]
    """
    The time when the user experience has returned to expected levels. Null if the alert is still open.
    """
    sepa_debit_delayed: Optional[SepaDebitDelayed]
    """
    Populated when type is sepa_debit_delayed.
    """
    severity: Literal["critical", "non_critical"]
    """
    The severity of the alert.
    """
    started_at: str
    """
    The time when impact on the user experience was first detected.
    """
    status: Literal["open", "resolved"]
    """
    The current status of the alert.
    """
    summary: str
    """
    A short description of the alert.
    """
    traffic_volume_drop: Optional[TrafficVolumeDrop]
    """
    Populated when type is traffic_volume_drop.
    """
    type: Literal[
        "api_error",
        "api_latency",
        "authorization_rate_drop",
        "elements_error",
        "event_generation_failure",
        "fraud_rate",
        "invoice_count_dropped",
        "issuing_authorization_request_errors",
        "issuing_authorization_request_timeout",
        "meter_event_summaries_delayed",
        "metronome_notification_latency",
        "payment_method_error",
        "sepa_debit_delayed",
        "traffic_volume_drop",
        "webhook_latency",
    ]
    """
    The type of the alert. Determines which sub-hash field is populated.
    """
    webhook_latency: Optional[WebhookLatency]
    """
    Populated when type is webhook_latency.
    """
    _inner_class_types = {
        "api_error": ApiError,
        "api_latency": ApiLatency,
        "authorization_rate_drop": AuthorizationRateDrop,
        "documentation_links": DocumentationLink,
        "elements_error": ElementsError,
        "event_generation_failure": EventGenerationFailure,
        "fraud_rate": FraudRate,
        "invoice_count_dropped": InvoiceCountDropped,
        "issuing_authorization_request_errors": IssuingAuthorizationRequestErrors,
        "issuing_authorization_request_timeout": IssuingAuthorizationRequestTimeout,
        "meter_event_summaries_delayed": MeterEventSummariesDelayed,
        "metronome_notification_latency": MetronomeNotificationLatency,
        "payment_method_error": PaymentMethodError,
        "sepa_debit_delayed": SepaDebitDelayed,
        "traffic_volume_drop": TrafficVolumeDrop,
        "webhook_latency": WebhookLatency,
    }



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from importlib import import_module
from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:
    from stripe.identity._blocklist_entry import (
        BlocklistEntry as BlocklistEntry,
    )
    from stripe.identity._blocklist_entry_service import (
        BlocklistEntryService as BlocklistEntryService,
    )
    from stripe.identity._verification_report import (
        VerificationReport as VerificationReport,
    )
    from stripe.identity._verification_report_service import (
        VerificationReportService as VerificationReportService,
    )
    from stripe.identity._verification_session import (
        VerificationSession as VerificationSession,
    )
    from stripe.identity._verification_session_service import (
        VerificationSessionService as VerificationSessionService,
    )

# name -> (import_target, is_submodule)
_import_map = {
    "BlocklistEntry": ("stripe.identity._blocklist_entry", False),
    "BlocklistEntryService": (
        "stripe.identity._blocklist_entry_service",
        False,
    ),
    "VerificationReport": ("stripe.identity._verification_report", False),
    "VerificationReportService": (
        "stripe.identity._verification_report_service",
        False,
    ),
    "VerificationSession": ("stripe.identity._verification_session", False),
    "VerificationSessionService": (
        "stripe.identity._verification_session_service",
        False,
    ),
}
if not TYPE_CHECKING:

    def __getattr__(name):
        try:
            target, is_submodule = _import_map[name]
            module = import_module(target)
            if is_submodule:
                return module

            return getattr(
                module,
                name,
            )
        except KeyError:
            raise AttributeError()



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from importlib import import_module
from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:
    from stripe.params.v2 import (
        billing as billing,
        core as core,
        extend as extend,
        iam as iam,
        money_management as money_management,
        network as network,
        orchestrated_commerce as orchestrated_commerce,
        payments as payments,
        reporting as reporting,
        signals as signals,
        tax as tax,
        test_helpers as test_helpers,
    )

# name -> (import_target, is_submodule)
_import_map = {
    "billing": ("stripe.params.v2.billing", True),
    "core": ("stripe.params.v2.core", True),
    "extend": ("stripe.params.v2.extend", True),
    "iam": ("stripe.params.v2.iam", True),
    "money_management": ("stripe.params.v2.money_management", True),
    "network": ("stripe.params.v2.network", True),
    "orchestrated_commerce": ("stripe.params.v2.orchestrated_commerce", True),
    "payments": ("stripe.params.v2.payments", True),
    "reporting": ("stripe.params.v2.reporting", True),
    "signals": ("stripe.params.v2.signals", True),
    "tax": ("stripe.params.v2.tax", True),
    "test_helpers": ("stripe.params.v2.test_helpers", True),
}
if not TYPE_CHECKING:

    def __getattr__(name):
        try:
            target, is_submodule = _import_map[name]
            module = import_module(target)
            if is_submodule:
                return module

            return getattr(
                module,
                name,
            )
        except KeyError:
            raise AttributeError()



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from stripe._stripe_object import UntypedStripeObject
from typing import Dict, List, Union
from typing_extensions import Literal, NotRequired, TypedDict


class PaymentIntentDecrementAuthorizationParams(RequestOptions):
    amount: int
    """
    The updated total amount that you intend to collect from the cardholder. This amount must be smaller than the currently authorized amount and greater than the already captured amount.
    """
    amount_details: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetails"
    ]
    """
    Provides industry-specific information about the amount.
    """
    application_fee_amount: NotRequired[int]
    """
    The amount of the application fee (if any) that will be requested to be applied to the payment and transferred to the application owner's Stripe account. The amount of the application fee collected will be capped at the total amount captured. For more information, see the PaymentIntents [use case for connected accounts](https://docs.stripe.com/payments/connected-accounts).
    """
    description: NotRequired[str]
    """
    An arbitrary string attached to the object. Often useful for displaying to users.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    hooks: NotRequired["PaymentIntentDecrementAuthorizationParamsHooks"]
    """
    Automations to be run during the PaymentIntent lifecycle
    """
    metadata: NotRequired["Dict[str, str]|UntypedStripeObject[str]"]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format. Individual keys can be unset by posting an empty value to them. All keys can be unset by posting an empty value to `metadata`.
    """
    payment_details: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsPaymentDetails"
    ]
    """
    Provides industry-specific information about the charge.
    """
    transfer_data: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsTransferData"
    ]
    """
    The parameters used to automatically create a transfer after the payment is captured.
    Learn more about the [use case for connected accounts](https://docs.stripe.com/payments/connected-accounts).
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetails(TypedDict):
    discount_amount: NotRequired["Literal['']|int"]
    """
    The total discount applied on the transaction represented in the [smallest currency unit](https://docs.stripe.com/currencies#zero-decimal). An integer greater than 0.

    This field is mutually exclusive with the `amount_details[line_items][#][discount_amount]` field.
    """
    enforce_arithmetic_validation: NotRequired[bool]
    """
    Set to `false` to return arithmetic validation errors in the response without failing the request. Use this when you want the operation to proceed regardless of arithmetic errors in the line item data.

    Omit or set to `true` to immediately return a 400 error when arithmetic validation fails. Use this for strict validation that prevents processing with line item data that has arithmetic inconsistencies.

    For card payments, Stripe doesn't send line item data to card networks if there's an arithmetic validation error.
    """
    line_items: NotRequired[
        "Literal['']|List[PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItem]"
    ]
    """
    A list of line items, each containing information about a product in the PaymentIntent. There is a maximum of 200 line items.
    """
    shipping: NotRequired[
        "Literal['']|PaymentIntentDecrementAuthorizationParamsAmountDetailsShipping"
    ]
    """
    Contains information about the shipping portion of the amount.
    """
    surcharge: NotRequired[
        "Literal['']|PaymentIntentDecrementAuthorizationParamsAmountDetailsSurcharge"
    ]
    """
    Contains information about the surcharge portion of the amount.
    """
    tax: NotRequired[
        "Literal['']|PaymentIntentDecrementAuthorizationParamsAmountDetailsTax"
    ]
    """
    Contains information about the tax portion of the amount.
    """
    tip: NotRequired[
        "Literal['']|PaymentIntentDecrementAuthorizationParamsAmountDetailsTip"
    ]
    """
    Contains information about the tip portion of the amount.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItem(
    TypedDict
):
    discount_amount: NotRequired[int]
    """
    The discount applied on this line item represented in the [smallest currency unit](https://docs.stripe.com/currencies#zero-decimal). An integer greater than 0.

    This field is mutually exclusive with the `amount_details[discount_amount]` field.
    """
    payment_method_options: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptions"
    ]
    """
    Payment method-specific information for line items.
    """
    product_code: NotRequired[str]
    """
    The product code of the line item, such as an SKU. Required for L3 rates. At most 12 characters long.
    """
    product_name: str
    """
    The product name of the line item. Required for L3 rates. At most 1024 characters long.

    For Cards, this field is truncated to 26 alphanumeric characters before being sent to the card networks. For PayPal, this field is truncated to 127 characters.
    """
    quantity: int
    """
    The quantity of items. Required for L3 rates. An integer greater than 0.
    """
    quantity_precision: NotRequired[int]
    """
    The number of decimal places implied in the quantity. For example, if quantity is 10000 and quantity_precision is 2, the actual quantity is 100.00. Defaults to 0 if not provided.
    """
    tax: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemTax"
    ]
    """
    Contains information about the tax on the item.
    """
    unit_cost: int
    """
    The unit cost of the line item represented in the [smallest currency unit](https://docs.stripe.com/currencies#zero-decimal). Required for L3 rates. An integer greater than or equal to 0.
    """
    unit_of_measure: NotRequired[str]
    """
    A unit of measure for the line item, such as gallons, feet, meters, etc.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptions(
    TypedDict,
):
    card: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsCard"
    ]
    """
    This sub-hash contains line item details that are specific to the `card` payment method.
    """
    card_present: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsCardPresent"
    ]
    """
    This sub-hash contains line item details that are specific to the `card_present` payment method.
    """
    klarna: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsKlarna"
    ]
    """
    This sub-hash contains line item details that are specific to the `klarna` payment method.
    """
    paypal: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsPaypal"
    ]
    """
    This sub-hash contains line item details that are specific to the `paypal` payment method.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsCard(
    TypedDict,
):
    commodity_code: NotRequired[str]
    """
    Identifier that categorizes the items being purchased using a standardized commodity scheme such as (but not limited to) UNSPSC, NAICS, NAPCS, and so on.
    """
    fleet_data: NotRequired[
        "PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsCardFleetData"
    ]
    """
    Fleet data for this line item.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsCardFleetData(
    TypedDict,
):
    product_type: Literal[
        "air_conditioning_service",
        "alcohol",
        "aviation_fuel_premium",
        "aviation_fuel_regular",
        "car_care_detailing",
        "compressed_natural_gas",
        "deli",
        "food_service",
        "green_gasoline_mid_plus",
        "green_gasoline_premium_super",
        "green_gasoline_regular",
        "grocery",
        "liquid_natural_gas",
        "liquid_propane_gas",
        "lodging",
        "marine_diesel",
        "marine_fuel",
        "merchandise",
        "mid_plus",
        "mid_plus_ethanol",
        "miscellaneous_aviation_products_services",
        "miscellaneous_fuel",
        "miscellaneous_marine_products_services",
        "miscellaneous_vehicle_products_services",
        "packaged_beverage",
        "premium_diesel",
        "premium_super",
        "premium_super_ethanol",
        "preventative_maintenance",
        "regular",
        "regular_diesel",
        "regular_ethanol",
        "repairs",
        "self_service_car_wash",
        "shower",
        "store_service",
        "tobacco",
        "vehicle_accessories",
        "vehicle_parking",
        "vehicle_parts",
        "wash_out",
    ]
    """
    The type of product being purchased at this line item.
    """
    service_type: NotRequired[
        Literal[
            "full_service",
            "high_speed_diesel",
            "non_fuel_only",
            "self_service",
        ]
    ]
    """
    The type of service received at the acceptor location.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsCardPresent(
    TypedDict,
):
    commodity_code: NotRequired[str]
    """
    Identifier that categorizes the items being purchased using a standardized commodity scheme such as (but not limited to) UNSPSC, NAICS, NAPCS, and so on.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsKlarna(
    TypedDict,
):
    image_url: NotRequired[str]
    """
    URL to an image for the product. Max length, 4096 characters.
    """
    product_url: NotRequired[str]
    """
    URL to the product page. Max length, 4096 characters.
    """
    reference: NotRequired[str]
    """
    Unique reference for this line item to correlate it with your system's internal records. The field is displayed in the Klarna Consumer App if passed.
    """
    subscription_reference: NotRequired[str]
    """
    Reference for the subscription this line item is for.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemPaymentMethodOptionsPaypal(
    TypedDict,
):
    category: NotRequired[
        Literal["digital_goods", "donation", "physical_goods"]
    ]
    """
    Type of the line item.
    """
    description: NotRequired[str]
    """
    Description of the line item.
    """
    sold_by: NotRequired[str]
    """
    The Stripe account ID of the connected account that sells the item.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsLineItemTax(
    TypedDict,
):
    total_tax_amount: int
    """
    The total amount of tax on a single line item represented in the [smallest currency unit](https://docs.stripe.com/currencies#zero-decimal). Required for L3 rates. An integer greater than or equal to 0.

    This field is mutually exclusive with the `amount_details[tax][total_tax_amount]` field.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsShipping(
    TypedDict
):
    amount: NotRequired["Literal['']|int"]
    """
    If a physical good is being shipped, the cost of shipping represented in the [smallest currency unit](https://docs.stripe.com/currencies#zero-decimal). An integer greater than or equal to 0.
    """
    from_postal_code: NotRequired["Literal['']|str"]
    """
    If a physical good is being shipped, the postal code of where it is being shipped from. At most 10 alphanumeric characters long, hyphens and spaces are allowed.
    """
    to_postal_code: NotRequired["Literal['']|str"]
    """
    If a physical good is being shipped, the postal code of where it is being shipped to. At most 10 alphanumeric characters long, hyphens and spaces are allowed.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsSurcharge(
    TypedDict,
):
    amount: NotRequired["Literal['']|int"]
    """
    Portion of the amount that corresponds to a surcharge.
    """
    enforce_validation: NotRequired[
        "Literal['']|Literal['automatic', 'disabled', 'enabled']"
    ]
    """
    Indicate whether to enforce validations on the surcharge amount.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsTax(TypedDict):
    total_tax_amount: int
    """
    The total amount of tax on the transaction represented in the [smallest currency unit](https://docs.stripe.com/currencies#zero-decimal). Required for L2 rates. An integer greater than or equal to 0.

    This field is mutually exclusive with the `amount_details[line_items][#][tax][total_tax_amount]` field.
    """


class PaymentIntentDecrementAuthorizationParamsAmountDetailsTip(TypedDict):
    amount: NotRequired["Literal['']|int"]
    """
    Portion of the amount that corresponds to a tip.
    """


class PaymentIntentDecrementAuthorizationParamsHooks(TypedDict):
    inputs: NotRequired["PaymentIntentDecrementAuthorizationParamsHooksInputs"]
    """
    Arguments passed in automations
    """


class PaymentIntentDecrementAuthorizationParamsHooksInputs(TypedDict):
    tax: NotRequired["PaymentIntentDecrementAuthorizationParamsHooksInputsTax"]
    """
    Tax arguments for automations
    """


class PaymentIntentDecrementAuthorizationParamsHooksInputsTax(TypedDict):
    calculation: Union[Literal[""], str]
    """
    The [TaxCalculation](https://docs.stripe.com/api/tax/calculations) id
    """


class PaymentIntentDecrementAuthorizationParamsPaymentDetails(TypedDict):
    customer_reference: NotRequired["Literal['']|str"]
    """
    A unique value to identify the customer. This field is available only for card payments.

    This field is truncated to 25 alphanumeric characters, excluding spaces, before being sent to card networks.
    """
    order_reference: NotRequired["Literal['']|str"]
    """
    A unique value assigned by the business to identify the transaction. Required for L2 and L3 rates.

    For Cards, this field is truncated to 25 alphanumeric characters, excluding spaces, before being sent to card networks. For Klarna, this field is truncated to 255 characters and is visible to customers when they view the order in the Klarna app.
    """


class PaymentIntentDecrementAuthorizationParamsTransferData(TypedDict):
    amount: NotRequired[int]
    """
    The amount that will be transferred automatically when a charge succeeds.
    """



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._createable_api_resource import CreateableAPIResource
from stripe._expandable_field import ExpandableField
from stripe._list_object import ListObject
from stripe._listable_api_resource import ListableAPIResource
from stripe._stripe_object import StripeObject, UntypedStripeObject
from stripe._updateable_api_resource import UpdateableAPIResource
from stripe._util import sanitize_id
from typing import ClassVar, List, Optional, cast
from typing_extensions import Literal, Unpack, TYPE_CHECKING

if TYPE_CHECKING:
    from stripe._application import Application
    from stripe.params.billing_portal._configuration_create_params import (
        ConfigurationCreateParams,
    )
    from stripe.params.billing_portal._configuration_list_params import (
        ConfigurationListParams,
    )
    from stripe.params.billing_portal._configuration_modify_params import (
        ConfigurationModifyParams,
    )
    from stripe.params.billing_portal._configuration_retrieve_params import (
        ConfigurationRetrieveParams,
    )


class Configuration(
    CreateableAPIResource["Configuration"],
    ListableAPIResource["Configuration"],
    UpdateableAPIResource["Configuration"],
):
    """
    A portal configuration describes the functionality and behavior you embed in a portal session. Related guide: [Configure the customer portal](https://docs.stripe.com/customer-management/configure-portal).
    """

    OBJECT_NAME: ClassVar[Literal["billing_portal.configuration"]] = (
        "billing_portal.configuration"
    )

    class BusinessProfile(StripeObject):
        headline: Optional[str]
        """
        The messaging shown to customers in the portal.
        """
        privacy_policy_url: Optional[str]
        """
        A link to the business's publicly available privacy policy.
        """
        terms_of_service_url: Optional[str]
        """
        A link to the business's publicly available terms of service.
        """

    class Features(StripeObject):
        class CustomerUpdate(StripeObject):
            allowed_updates: List[
                Literal[
                    "address", "email", "name", "phone", "shipping", "tax_id"
                ]
            ]
            """
            The types of customer updates that are supported. When empty, customers are not updateable.
            """
            enabled: bool
            """
            Whether the feature is enabled.
            """

        class InvoiceHistory(StripeObject):
            enabled: bool
            """
            Whether the feature is enabled.
            """

        class PaymentMethodUpdate(StripeObject):
            enabled: bool
            """
            Whether the feature is enabled.
            """
            payment_method_configuration: Optional[str]
            """
            The [Payment Method Configuration](https://docs.stripe.com/api/payment_method_configurations) to use for this portal session. When specified, customers will be able to update their payment method to one of the options specified by the payment method configuration. If not set, the default payment method configuration is used.
            """

        class SubscriptionCancel(StripeObject):
            class CancellationReason(StripeObject):
                enabled: bool
                """
                Whether the feature is enabled.
                """
                options: List[
                    Literal[
                        "customer_service",
                        "low_quality",
                        "missing_features",
                        "other",
                        "switched_service",
                        "too_complex",
                        "too_expensive",
                        "unused",
                    ]
                ]
                """
                Which cancellation reasons will be given as options to the customer.
                """

            cancellation_reason: CancellationReason
            enabled: bool
            """
            Whether the feature is enabled.
            """
            mode: Literal["at_period_end", "immediately"]
            """
            Whether to cancel subscriptions immediately or at the end of the billing period.
            """
            proration_behavior: Literal[
                "always_invoice", "create_prorations", "none"
            ]
            """
            Whether to create prorations when canceling subscriptions. Possible values are `none` and `create_prorations`.
            """
            _inner_class_types = {"cancellation_reason": CancellationReason}

        class SubscriptionUpdate(StripeObject):
            class Product(StripeObject):
                class AdjustableQuantity(StripeObject):
                    enabled: bool
                    """
                    If true, the quantity can be adjusted to any non-negative integer.
                    """
                    maximum: Optional[int]
                    """
                    The maximum quantity that can be set for the product.
                    """
                    minimum: int
                    """
                    The minimum quantity that can be set for the product.
                    """

                adjustable_quantity: AdjustableQuantity
                prices: List[str]
                """
                The list of price IDs which, when subscribed to, a subscription can be updated.
                """
                product: str
                """
                The product ID.
                """
                _inner_class_types = {
                    "adjustable_quantity": AdjustableQuantity
                }

            class ScheduleAtPeriodEnd(StripeObject):
                class Condition(StripeObject):
                    type: Literal[
                        "decreasing_item_amount", "shortening_interval"
                    ]
                    """
                    The type of condition.
                    """

                conditions: List[Condition]
                """
                List of conditions. When any condition is true, an update will be scheduled at the end of the current period.
                """
                _inner_class_types = {"conditions": Condition}

            billing_cycle_anchor: Optional[Literal["now", "unchanged"]]
            """
            Determines the value to use for the billing cycle anchor on subscription updates. Valid values are `now` or `unchanged`, and the default value is `unchanged`. Setting the value to `now` resets the subscription's billing cycle anchor to the current time (in UTC). For more information, see the billing cycle [documentation](https://docs.stripe.com/billing/subscriptions/billing-cycle).
            """
            default_allowed_updates: List[
                Literal["price", "promotion_code", "quantity"]
            ]
            """
            The types of subscription updates that are supported for items listed in the `products` attribute. When empty, subscriptions are not updateable.
            """
            enabled: bool
            """
            Whether the feature is enabled.
            """
            products: Optional[List[Product]]
            """
            The list of up to 10 products that support subscription updates.
            """
            proration_behavior: Literal[
                "always_invoice", "create_prorations", "none"
            ]
            """
            Determines how to handle prorations resulting from subscription updates. Valid values are `none`, `create_prorations`, and `always_invoice`. Defaults to a value of `none` if you don't set it during creation.
            """
            schedule_at_period_end: ScheduleAtPeriodEnd
            trial_update_behavior: Literal["continue_trial", "end_trial"]
            """
            Determines how handle updates to trialing subscriptions. Valid values are `end_trial` and `continue_trial`. Defaults to a value of `end_trial` if you don't set it during creation.
            """
            _inner_class_types = {
                "products": Product,
                "schedule_at_period_end": ScheduleAtPeriodEnd,
            }

        customer_update: CustomerUpdate
        invoice_history: InvoiceHistory
        payment_method_update: PaymentMethodUpdate
        subscription_cancel: SubscriptionCancel
        subscription_update: SubscriptionUpdate
        _inner_class_types = {
            "customer_update": CustomerUpdate,
            "invoice_history": InvoiceHistory,
            "payment_method_update": PaymentMethodUpdate,
            "subscription_cancel": SubscriptionCancel,
            "subscription_update": SubscriptionUpdate,
        }

    class LoginPage(StripeObject):
        enabled: bool
        """
        If `true`, a shareable `url` will be generated that will take your customers to a hosted login page for the customer portal.

        If `false`, the previously generated `url`, if any, will be deactivated.
        """
        url: Optional[str]
        """
        A shareable URL to the hosted portal login page. Your customers will be able to log in with their [email](https://docs.stripe.com/api/customers/object#customer_object-email) and receive a link to their customer portal.
        """

    active: bool
    """
    Whether the configuration is active and can be used to create portal sessions.
    """
    application: Optional[ExpandableField["Application"]]
    """
    ID of the Connect Application that created the configuration.
    """
    business_profile: BusinessProfile
    created: int
    """
    Time at which the object was created. Measured in seconds since the Unix epoch.
    """
    default_return_url: Optional[str]
    """
    The default URL to redirect customers to when they click on the portal's link to return to your website. This can be [overriden](https://docs.stripe.com/api/customer_portal/sessions/create#create_portal_session-return_url) when creating the session.
    """
    features: Features
    id: str
    """
    Unique identifier for the object.
    """
    is_default: bool
    """
    Whether the configuration is the default. If `true`, this configuration can be managed in the Dashboard and portal sessions will use this configuration unless it is overriden when creating the session.
    """
    livemode: bool
    """
    If the object exists in live mode, the value is `true`. If the object exists in test mode, the value is `false`.
    """
    login_page: LoginPage
    metadata: Optional[UntypedStripeObject[str]]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format.
    """
    name: Optional[str]
    """
    The name of the configuration.
    """
    object: Literal["billing_portal.configuration"]
    """
    String representing the object's type. Objects of the same type share the same value.
    """
    updated: int
    """
    Time at which the object was last updated. Measured in seconds since the Unix epoch.
    """

    @classmethod
    def create(
        cls, **params: Unpack["ConfigurationCreateParams"]
    ) -> "Configuration":
        """
        Creates a configuration that describes the functionality and behavior of a PortalSession
        """
        return cast(
            "Configuration",
            cls._static_request(
                "post",
                cls.class_url(),
                params=params,
            ),
        )

    @classmethod
    async def create_async(
        cls, **params: Unpack["ConfigurationCreateParams"]
    ) -> "Configuration":
        """
        Creates a configuration that describes the functionality and behavior of a PortalSession
        """
        return cast(
            "Configuration",
            await cls._static_request_async(
                "post",
                cls.class_url(),
                params=params,
            ),
        )

    @classmethod
    def list(
        cls, **params: Unpack["ConfigurationListParams"]
    ) -> ListObject["Configuration"]:
        """
        Returns a list of configurations that describe the functionality of the customer portal.
        """
        result = cls._static_request(
            "get",
            cls.class_url(),
            params=params,
        )
        if not isinstance(result, ListObject):
            raise TypeError(
                "Expected list object from API, got %s"
                % (type(result).__name__)
            )

        return result

    @classmethod
    async def list_async(
        cls, **params: Unpack["ConfigurationListParams"]
    ) -> ListObject["Configuration"]:
        """
        Returns a list of configurations that describe the functionality of the customer portal.
        """
        result = await cls._static_request_async(
            "get",
            cls.class_url(),
            params=params,
        )
        if not isinstance(result, ListObject):
            raise TypeError(
                "Expected list object from API, got %s"
                % (type(result).__name__)
            )

        return result

    @classmethod
    def modify(
        cls, id: str, **params: Unpack["ConfigurationModifyParams"]
    ) -> "Configuration":
        """
        Updates a configuration that describes the functionality of the customer portal.
        """
        url = "%s/%s" % (cls.class_url(), sanitize_id(id))
        return cast(
            "Configuration",
            cls._static_request(
                "post",
                url,
                params=params,
            ),
        )

    @classmethod
    async def modify_async(
        cls, id: str, **params: Unpack["ConfigurationModifyParams"]
    ) -> "Configuration":
        """
        Updates a configuration that describes the functionality of the customer portal.
        """
        url = "%s/%s" % (cls.class_url(), sanitize_id(id))
        return cast(
            "Configuration",
            await cls._static_request_async(
                "post",
                url,
                params=params,
            ),
        )

    @classmethod
    def retrieve(
        cls, id: str, **params: Unpack["ConfigurationRetrieveParams"]
    ) -> "Configuration":
        """
        Retrieves a configuration that describes the functionality of the customer portal.
        """
        instance = cls(id, **params)
        instance.refresh()
        return instance

    @classmethod
    async def retrieve_async(
        cls, id: str, **params: Unpack["ConfigurationRetrieveParams"]
    ) -> "Configuration":
        """
        Retrieves a configuration that describes the functionality of the customer portal.
        """
        instance = cls(id, **params)
        await instance.refresh_async()
        return instance

    _inner_class_types = {
        "business_profile": BusinessProfile,
        "features": Features,
        "login_page": LoginPage,
    }



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from typing import List
from typing_extensions import Literal, NotRequired, TypedDict


class RedactionJobCreateParams(RequestOptions):
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    objects: "RedactionJobCreateParamsObjects"
    """
    The objects to redact. These root objects and their related ones will be validated for redaction.
    """
    validation_behavior: NotRequired[Literal["error", "fix"]]
    """
    Determines the validation behavior of the job. Default is `error`.
    """


class RedactionJobCreateParamsObjects(TypedDict):
    charges: NotRequired[List[str]]
    checkout_sessions: NotRequired[List[str]]
    customers: NotRequired[List[str]]
    identity_verification_sessions: NotRequired[List[str]]
    invoices: NotRequired[List[str]]
    issuing_cardholders: NotRequired[List[str]]
    issuing_cards: NotRequired[List[str]]
    payment_intents: NotRequired[List[str]]
    radar_value_list_items: NotRequired[List[str]]
    setup_intents: NotRequired[List[str]]



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from stripe._stripe_object import UntypedStripeObject
from typing import Dict, List
from typing_extensions import Literal, NotRequired, TypedDict


class ReaderCollectInputsParams(RequestOptions):
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    inputs: List["ReaderCollectInputsParamsInput"]
    """
    List of inputs to be collected from the customer using the Reader. Maximum 5 inputs.
    """
    metadata: NotRequired["Dict[str, str]|UntypedStripeObject[str]"]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format. Individual keys can be unset by posting an empty value to them. All keys can be unset by posting an empty value to `metadata`.
    """


class ReaderCollectInputsParamsInput(TypedDict):
    custom_text: "ReaderCollectInputsParamsInputCustomText"
    """
    Customize the text which will be displayed while collecting this input
    """
    required: NotRequired[bool]
    """
    Indicate that this input is required, disabling the skip button
    """
    selection: NotRequired["ReaderCollectInputsParamsInputSelection"]
    """
    Options for the `selection` input
    """
    toggles: NotRequired[List["ReaderCollectInputsParamsInputToggle"]]
    """
    List of toggles to be displayed and customization for the toggles
    """
    type: Literal[
        "email", "numeric", "phone", "selection", "signature", "text"
    ]
    """
    The type of input to collect
    """


class ReaderCollectInputsParamsInputCustomText(TypedDict):
    description: NotRequired[str]
    """
    The description which will be displayed when collecting this input
    """
    skip_button: NotRequired[str]
    """
    Custom text for the skip button. Maximum 14 characters.
    """
    submit_button: NotRequired[str]
    """
    Custom text for the submit button. Maximum 30 characters.
    """
    title: str
    """
    The title which will be displayed when collecting this input
    """


class ReaderCollectInputsParamsInputSelection(TypedDict):
    choices: List["ReaderCollectInputsParamsInputSelectionChoice"]
    """
    List of choices for the `selection` input
    """


class ReaderCollectInputsParamsInputSelectionChoice(TypedDict):
    id: str
    """
    The unique identifier for this choice
    """
    style: NotRequired[Literal["primary", "secondary"]]
    """
    The style of the button which will be shown for this choice. Can be `primary` or `secondary`.
    """
    text: str
    """
    The text which will be shown on the button for this choice
    """


class ReaderCollectInputsParamsInputToggle(TypedDict):
    default_value: NotRequired[Literal["disabled", "enabled"]]
    """
    The default value of the toggle. Can be `enabled` or `disabled`.
    """
    description: NotRequired[str]
    """
    The description which will be displayed for the toggle. Maximum 50 characters. At least one of title or description must be provided.
    """
    title: NotRequired[str]
    """
    The title which will be displayed for the toggle. Maximum 50 characters. At least one of title or description must be provided.
    """



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._stripe_object import StripeObject, UntypedStripeObject
from typing import ClassVar, Optional
from typing_extensions import Literal


class BatchJob(StripeObject):
    """
    A batch job allows you to perform an API operation on a large set of records asynchronously.
    """

    OBJECT_NAME: ClassVar[Literal["v2.core.batch_job"]] = "v2.core.batch_job"

    class StatusDetails(StripeObject):
        class BatchFailed(StripeObject):
            error: str
            """
            Details about the `batch_job` failure.
            """

        class Canceled(StripeObject):
            class OutputFile(StripeObject):
                class DownloadUrl(StripeObject):
                    expires_at: Optional[str]
                    """
                    The time that the URL expires.
                    """
                    url: str
                    """
                    The URL that can be used for accessing the file.
                    """

                content_type: str
                """
                The content type of the file.
                """
                download_url: DownloadUrl
                """
                A pre-signed URL that allows secure, time-limited access to download the file.
                """
                size: int
                """
                The total size of the file in bytes.
                """
                _inner_class_types = {"download_url": DownloadUrl}
                _field_encodings = {"size": "int64_string"}

            failure_count: int
            """
            The total number of records that failed processing.
            """
            output_file: OutputFile
            """
            The output file details. If the `batch_job` is canceled, this is provided only if there is already output at this point.
            """
            success_count: int
            """
            The total number of records that were successfully processed.
            """
            _inner_class_types = {"output_file": OutputFile}
            _field_encodings = {
                "failure_count": "int64_string",
                "success_count": "int64_string",
            }

        class Complete(StripeObject):
            class OutputFile(StripeObject):
                class DownloadUrl(StripeObject):
                    expires_at: Optional[str]
                    """
                    The time that the URL expires.
                    """
                    url: str
                    """
                    The URL that can be used for accessing the file.
                    """

                content_type: str
                """
                The content type of the file.
                """
                download_url: DownloadUrl
                """
                A pre-signed URL that allows secure, time-limited access to download the file.
                """
                size: int
                """
                The total size of the file in bytes.
                """
                _inner_class_types = {"download_url": DownloadUrl}
                _field_encodings = {"size": "int64_string"}

            failure_count: int
            """
            The total number of records that failed processing.
            """
            output_file: OutputFile
            """
            The output file details. If the `batch_job` is canceled, this is provided only if there is already output at this point.
            """
            success_count: int
            """
            The total number of records that were successfully processed.
            """
            _inner_class_types = {"output_file": OutputFile}
            _field_encodings = {
                "failure_count": "int64_string",
                "success_count": "int64_string",
            }

        class InProgress(StripeObject):
            failure_count: int
            """
            The number of records that failed processing so far.
            """
            success_count: int
            """
            The number of records that were successfully processed so far.
            """
            _field_encodings = {
                "failure_count": "int64_string",
                "success_count": "int64_string",
            }

        class ReadyForUpload(StripeObject):
            class UploadUrl(StripeObject):
                expires_at: Optional[str]
                """
                The time that the URL expires.
                """
                url: str
                """
                The URL that can be used for accessing the file.
                """

            upload_url: UploadUrl
            """
            The upload file details.
            """
            _inner_class_types = {"upload_url": UploadUrl}

        class Timeout(StripeObject):
            class OutputFile(StripeObject):
                class DownloadUrl(StripeObject):
                    expires_at: Optional[str]
                    """
                    The time that the URL expires.
                    """
                    url: str
                    """
                    The URL that can be used for accessing the file.
                    """

                content_type: str
                """
                The content type of the file.
                """
                download_url: DownloadUrl
                """
                A pre-signed URL that allows secure, time-limited access to download the file.
                """
                size: int
                """
                The total size of the file in bytes.
                """
                _inner_class_types = {"download_url": DownloadUrl}
                _field_encodings = {"size": "int64_string"}

            failure_count: int
            """
            The total number of records that failed processing.
            """
            output_file: OutputFile
            """
            The output file details. If the `batch_job` is canceled, this is provided only if there is already output at this point.
            """
            success_count: int
            """
            The total number of records that were successfully processed.
            """
            _inner_class_types = {"output_file": OutputFile}
            _field_encodings = {
                "failure_count": "int64_string",
                "success_count": "int64_string",
            }

        class Validating(StripeObject):
            validated_count: int
            """
            The number of records that were validated. Note that there is no failure counter here;
            once we have any validation failures we give up.
            """
            _field_encodings = {"validated_count": "int64_string"}

        class ValidationFailed(StripeObject):
            class OutputFile(StripeObject):
                class DownloadUrl(StripeObject):
                    expires_at: Optional[str]
                    """
                    The time that the URL expires.
                    """
                    url: str
                    """
                    The URL that can be used for accessing the file.
                    """

                content_type: str
                """
                The content type of the file.
                """
                download_url: DownloadUrl
                """
                A pre-signed URL that allows secure, time-limited access to download the file.
                """
                size: int
                """
                The total size of the file in bytes.
                """
                _inner_class_types = {"download_url": DownloadUrl}
                _field_encodings = {"size": "int64_string"}

            failure_count: int
            """
            The total number of records that failed processing.
            """
            output_file: OutputFile
            """
            The output file details. If the `batch_job` is canceled, this is provided only if there is already output at this point.
            """
            success_count: int
            """
            The total number of records that were successfully processed.
            """
            _inner_class_types = {"output_file": OutputFile}
            _field_encodings = {
                "failure_count": "int64_string",
                "success_count": "int64_string",
            }

        batch_failed: Optional[BatchFailed]
        """
        Additional details for the `BATCH_FAILED` status of the `batch_job`.
        """
        canceled: Optional[Canceled]
        """
        Additional details for the `CANCELED` status of the `batch_job`.
        """
        complete: Optional[Complete]
        """
        Additional details for the `COMPLETE` status of the `batch_job`.
        """
        in_progress: Optional[InProgress]
        """
        Additional details for the `IN_PROGRESS` status of the `batch_job`.
        """
        ready_for_upload: Optional[ReadyForUpload]
        """
        Additional details for the `READY_FOR_UPLOAD` status of the `batch_job`.
        """
        timeout: Optional[Timeout]
        """
        Additional details for the `TIMEOUT` status of the `batch_job`.
        """
        validating: Optional[Validating]
        """
        Additional details for the `VALIDATING` status of the `batch_job`.
        """
        validation_failed: Optional[ValidationFailed]
        """
        Additional details for the `VALIDATION_FAILED` status of the `batch_job`.
        """
        _inner_class_types = {
            "batch_failed": BatchFailed,
            "canceled": Canceled,
            "complete": Complete,
            "in_progress": InProgress,
            "ready_for_upload": ReadyForUpload,
            "timeout": Timeout,
            "validating": Validating,
            "validation_failed": ValidationFailed,
        }

    created: str
    """
    Timestamp at which the `batch_job` was created.
    """
    id: str
    """
    Unique identifier for the `batch_job`.
    """
    livemode: bool
    """
    Has the value `true` if the object exists in live mode or the value `false` if the object exists in test mode.
    """
    metadata: UntypedStripeObject[str]
    """
    The metadata of the `batch_job`.
    """
    object: Literal["v2.core.batch_job"]
    """
    String representing the object's type. Objects of the same type share the same value of the object field.
    """
    skip_validation: bool
    """
    Whether validation runs before executing the `batch_job`.
    """
    status: Literal[
        "batch_failed",
        "canceled",
        "cancelling",
        "complete",
        "in_progress",
        "ready_for_upload",
        "timeout",
        "upload_timeout",
        "validating",
        "validation_failed",
    ]
    """
    The current status of the `batch_job`.
    """
    status_details: Optional[StatusDetails]
    """
    Additional details about the current state of the `batch_job`.
    """
    _inner_class_types = {"status_details": StatusDetails}



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from stripe._stripe_object import UntypedStripeObject
from typing import Dict, List
from typing_extensions import Literal, NotRequired, TypedDict


class OnrampSessionCreateParams(RequestOptions):
    customer_ip_address: NotRequired[str]
    """
    The IP address of the customer the platform intends to onramp.

    If the user's IP is in a region we can't support, we return an `HTTP 400` with an appropriate error code.

    We support IPv4 and IPv6 addresses. Geographic supportability is checked again later in the onramp flow, which provides a way to hide the onramp option from ineligible users for a better user experience.
    """
    destination_amount: NotRequired[str]
    """
    The default amount of crypto to exchange into.

    * When left null, a default value is computed if `source_amount`, `destination_currency`, and `destination_network` are set.
    * When set, both `destination_currency` and `destination_network` must also be set. All cryptocurrencies are supported to their full precisions (for example, 18 decimal places for `eth`). We validate and generate an error if the amount exceeds the supported precision based on the exchange currency. Setting `source_amount` is mutually exclusive with setting `destination_amount` (only one or the other is supported). Users can update the amount in the onramp UI.
    """
    destination_currencies: NotRequired[
        List[
            Literal["avax", "btc", "eth", "matic", "sol", "usdc", "wld", "xlm"]
        ]
    ]
    """
    The list of destination cryptocurrencies a user can choose from.

    * When left null, all supported cryptocurrencies are shown in the onramp UI subject to `destination_networks` if set.
    * When set, it must be a non-empty array where all values in the array are valid cryptocurrencies. You can use it to lock users to a specific cryptocurrency by passing a single value array. Users **cannot** override this parameter.
    """
    destination_currency: NotRequired[
        Literal["avax", "btc", "eth", "matic", "sol", "usdc", "wld", "xlm"]
    ]
    """
    The default destination cryptocurrency.

    * When left null, the first value of `destination_currencies` is selected.
    * When set, if `destination_currencies` is also set, the value of `destination_currency` must be present in that array. To lock a `destination_currency`, specify that value as the single value for `destination_currencies`. Users can select a different cryptocurrency in the onramp UI subject to `destination_currencies` if set.
    """
    destination_network: NotRequired[
        Literal[
            "avalanche",
            "base",
            "bitcoin",
            "ethereum",
            "optimism",
            "polygon",
            "solana",
            "stellar",
            "worldchain",
        ]
    ]
    """
    The default destination crypto network.

    * When left null, the first value of `destination_networks` is selected.
    * When set, if `destination_networks` is also set, the value of `destination_network` must be present in that array. To lock a `destination_network`, specify that value as the single value for `destination_networks`. Users can select a different network in the onramp UI subject to `destination_networks` if set.
    """
    destination_networks: NotRequired[
        List[
            Literal[
                "avalanche",
                "base",
                "bitcoin",
                "ethereum",
                "optimism",
                "polygon",
                "solana",
                "stellar",
                "worldchain",
            ]
        ]
    ]
    """
    The list of destination crypto networks user can choose from.

    * When left null, all supported crypto networks are shown in the onramp UI.
    * When set, it must be a non-empty array where values in the array are each a valid crypto network. It can be used to lock users to a specific network by passing a single value array. Users **cannot** override this parameter.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    kyc_details: NotRequired["OnrampSessionCreateParamsKycDetails"]
    """
    Pre-populate some of the required KYC information for the user if you've already collected it within your application.

    Related guide: [Using the API](https://docs.stripe.com/crypto/using-the-api#how-to-pre-populate-customer-information)
    """
    lock_wallet_address: NotRequired[bool]
    """
    Whether or not to lock the suggested wallet address. If destination tags are provided, this will also lock the destination tags.
    """
    metadata: NotRequired["Dict[str, str]|UntypedStripeObject[str]"]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format. Individual keys can be unset by posting an empty value to them. All keys can be unset by posting an empty value to `metadata`.
    """
    settlement_speed: NotRequired[Literal["instant", "standard"]]
    """
    Speed at which the cryptocurrency is delivered to the wallet
    One of:
     `instant` (default): crypto is delivered when payment is confirmed
     `standard`: crypto is delivered when payment settles
    """
    source_amount: NotRequired[str]
    """
    The default amount of fiat (in decimal) to exchange into crypto.

    * When left null, a default value is computed if `destination_amount` is set.
    * When set, setting `source_amount` is mutually exclusive with setting `destination_amount` (only one or the other is supported). We don't support fractional pennies. If fractional minor units of a currency are passed in, it generates an error. Users can update the value in the onramp UI.
    """
    source_currency: NotRequired[Literal["eur", "gbp", "usd"]]
    """
    The default source fiat currency for the onramp session.

    * When left null, a default currency is selected based on user locale.
    * When set, it must be one of the fiat currencies supported by onramp. Users can still select a different currency in the onramp UI.
    """
    wallet_addresses: NotRequired["OnrampSessionCreateParamsWalletAddresses"]
    """
    The end customer's crypto wallet address (for each network) to use for this transaction.

    * When left null, the user enters their wallet in the onramp UI.
    * When set, the platform must set either `destination_networks` or `destination_network` and we perform address validation. Users can still select a different wallet in the onramp UI.
    """


class OnrampSessionCreateParamsKycDetails(TypedDict):
    pass


class OnrampSessionCreateParamsWalletAddresses(TypedDict):
    destination_tags: NotRequired[
        "OnrampSessionCreateParamsWalletAddressesDestinationTags"
    ]
    """
    The end customer's crypto wallet destination tag (for each network) to use for this transaction. This only applies for tag-based assets such as XLM.

    * When left null, the user enters their wallet in the onramp UI.
    """


class OnrampSessionCreateParamsWalletAddressesDestinationTags(TypedDict):
    stellar: NotRequired[str]



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._createable_api_resource import CreateableAPIResource
from stripe._list_object import ListObject
from stripe._listable_api_resource import ListableAPIResource
from stripe._stripe_object import StripeObject, UntypedStripeObject
from stripe._util import class_method_variant, sanitize_id
from typing import ClassVar, List, Optional, cast, overload
from typing_extensions import Literal, Unpack, TYPE_CHECKING

if TYPE_CHECKING:
    from stripe.params.crypto._onramp_session_checkout_params import (
        OnrampSessionCheckoutParams,
    )
    from stripe.params.crypto._onramp_session_create_params import (
        OnrampSessionCreateParams,
    )
    from stripe.params.crypto._onramp_session_list_params import (
        OnrampSessionListParams,
    )
    from stripe.params.crypto._onramp_session_quote_params import (
        OnrampSessionQuoteParams,
    )
    from stripe.params.crypto._onramp_session_retrieve_params import (
        OnrampSessionRetrieveParams,
    )


class OnrampSession(
    CreateableAPIResource["OnrampSession"],
    ListableAPIResource["OnrampSession"],
):
    """
    A Crypto Onramp Session represents your customer's session as they purchase cryptocurrency through Stripe. Once payment is successful, Stripe will fulfill the delivery of cryptocurrency to your user's wallet and contain a reference to the crypto transaction ID.

    You can create an onramp session on your server and embed the widget on your frontend. Alternatively, you can redirect your users to the standalone hosted onramp.

    Related guide: [Integrate the onramp](https://docs.stripe.com/crypto/integrate-the-onramp)
    """

    OBJECT_NAME: ClassVar[Literal["crypto.onramp_session"]] = (
        "crypto.onramp_session"
    )

    class TransactionDetails(StripeObject):
        class Fees(StripeObject):
            network_fee_amount: Optional[str]
            """
            The cost associated with moving crypto from Stripe to the end consumer's wallet. e.g: for ETH, this is called 'gas fee', for BTC this is a 'miner's fee'.
            """
            transaction_fee_amount: Optional[str]
            """
            Fee for processing the transaction.
            """

        class WalletAddresses(StripeObject):
            class DestinationTags(StripeObject):
                stellar: Optional[str]
                """
                A stellar destination tag
                """

            avalanche: Optional[str]
            """
            An avalanche address
            """
            base_network: Optional[str]
            """
            A base address
            """
            bitcoin: Optional[str]
            """
            A bitcoin address
            """
            destination_tags: Optional[DestinationTags]
            """
            The end customer's crypto wallet destination tag (for each network) to use for this transaction.
            """
            ethereum: Optional[str]
            """
            An ethereum address
            """
            optimism: Optional[str]
            """
            An optimism address
            """
            polygon: Optional[str]
            """
            A polygon address
            """
            solana: Optional[str]
            """
            A solana address
            """
            stellar: Optional[str]
            """
            A stellar address
            """
            worldchain: Optional[str]
            """
            A worldchain address
            """
            _inner_class_types = {"destination_tags": DestinationTags}

        destination_amount: Optional[str]
        """
        The amount of crypto the customer will get deposited into their wallet
        """
        destination_currencies: Optional[
            List[
                Literal[
                    "avax", "btc", "eth", "matic", "sol", "usdc", "wld", "xlm"
                ]
            ]
        ]
        """
        If a platform wants to lock the currencies an session will support, they can add supported currencies to this array. If left null, the experience will allow selection of all supported destination currencies.
        """
        destination_currency: Optional[
            Literal["avax", "btc", "eth", "matic", "sol", "usdc", "wld", "xlm"]
        ]
        """
        The selected `destination_currency` to convert the `source` to. This should be a crypto currency code. If `destination_currencies` is set, it must be a value in that array.
        """
        destination_network: Optional[
            Literal[
                "avalanche",
                "base",
                "bitcoin",
                "ethereum",
                "optimism",
                "polygon",
                "solana",
                "stellar",
                "worldchain",
            ]
        ]
        """
        The specific crypto network the `destination_currency` is settled on. If `destination_networks` is set, it must be a value in that array.
        """
        destination_networks: Optional[
            List[
                Literal[
                    "avalanche",
                    "base",
                    "bitcoin",
                    "ethereum",
                    "optimism",
                    "polygon",
                    "solana",
                    "stellar",
                    "worldchain",
                ]
            ]
        ]
        """
        If a platform wants to lock the supported networks, they can do so through this array. If left null, the experience will allow selection of all supported networks.
        """
        fees: Optional[Fees]
        """
        Details about the fees associated with this transaction
        """
        lock_wallet_address: Optional[bool]
        """
        Whether or not to lock the suggested wallet address.
        """
        settlement_speed: Optional[Literal["instant", "standard"]]
        """
        Speed at which the cryptocurrency is delivered to the wallet
        One of:
         `instant` (default): crypto is delivered when payment is confirmed
         `standard`: crypto is delivered when payment settles
        """
        source_amount: Optional[str]
        """
        The amount of fiat we intend to onramp - excluding fees
        """
        source_currency: Optional[Literal["eur", "gbp", "usd"]]
        """
        A fiat currency code
        """
        transaction_id: Optional[str]
        """
        The on-chain transaction hash (also referred to as transaction ID or tx_hash) of the transaction that was sent to the customer's wallet. The format varies by chain (e.g. `0xc257...1a95` on Ethereum, `5UB1...v3xZ` on Solana, or `a1b2...bf00` on Bitcoin). This will only be set if the session reaches `status=fulfillment_complete` and we've transferred the crypto successfully to the external wallet.
        """
        wallet_address: Optional[str]
        """
        The consumer's wallet address (where crypto will be sent to)
        """
        wallet_addresses: Optional[WalletAddresses]
        """
        The end customer's crypto wallet address (for each network) to use for this transaction.
        """
        _inner_class_types = {
            "fees": Fees,
            "wallet_addresses": WalletAddresses,
        }

    client_secret: str
    """
    A client secret that can be used to drive a single session using our embedded widget.

    Related guide: [Set up an onramp integration](https://docs.stripe.com/crypto/integrate-the-onramp)
    """
    created: int
    """
    Time at which the object was created. Measured in seconds since the Unix epoch.
    """
    id: str
    """
    Unique identifier for the object.
    """
    kyc_details_provided: bool
    """
    Has the value `true` if any user kyc details were provided during the creation of the onramp session. Otherwise, has the value `false`.
    """
    livemode: bool
    """
    If the object exists in live mode, the value is `true`. If the object exists in test mode, the value is `false`.
    """
    metadata: Optional[UntypedStripeObject[str]]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format.
    """
    object: Literal["crypto.onramp_session"]
    """
    String representing the object's type. Objects of the same type share the same value.
    """
    redirect_url: Optional[str]
    """
    Redirect your users to the URL for a prebuilt frontend integration of the crypto onramp on the standalone hosted onramp.

    Related guide: [Mint a session with a redirect url](https://docs.stripe.com/crypto/standalone-hosted-onramp#mint-a-session-with-a-redirect-url)
    """
    status: str
    """
    The status of the Onramp Session. One of = `{initialized, rejected, requires_payment, fulfillment_processing, fulfillment_complete}`
    """
    transaction_details: TransactionDetails

    @classmethod
    def _cls_checkout(
        cls, id: str, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        return cast(
            "OnrampSession",
            cls._static_request(
                "post",
                "/v1/crypto/onramp_sessions/{id}/checkout".format(
                    id=sanitize_id(id)
                ),
                params=params,
            ),
        )

    @overload
    @staticmethod
    def checkout(
        id: str, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        ...

    @overload
    def checkout(
        self, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        ...

    @class_method_variant("_cls_checkout")
    def checkout(  # pyright: ignore[reportGeneralTypeIssues]
        self, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        return cast(
            "OnrampSession",
            self._request(
                "post",
                "/v1/crypto/onramp_sessions/{id}/checkout".format(
                    id=sanitize_id(self._data.get("id"))
                ),
                params=params,
            ),
        )

    @classmethod
    async def _cls_checkout_async(
        cls, id: str, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        return cast(
            "OnrampSession",
            await cls._static_request_async(
                "post",
                "/v1/crypto/onramp_sessions/{id}/checkout".format(
                    id=sanitize_id(id)
                ),
                params=params,
            ),
        )

    @overload
    @staticmethod
    async def checkout_async(
        id: str, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        ...

    @overload
    async def checkout_async(
        self, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        ...

    @class_method_variant("_cls_checkout_async")
    async def checkout_async(  # pyright: ignore[reportGeneralTypeIssues]
        self, **params: Unpack["OnrampSessionCheckoutParams"]
    ) -> "OnrampSession":
        """
        Completes a headless CryptoOnrampSession.

        This method will attempt to confirm the payment and execute the quote to deliver the crypto to the customer.
        """
        return cast(
            "OnrampSession",
            await self._request_async(
                "post",
                "/v1/crypto/onramp_sessions/{id}/checkout".format(
                    id=sanitize_id(self._data.get("id"))
                ),
                params=params,
            ),
        )

    @classmethod
    def create(
        cls, **params: Unpack["OnrampSessionCreateParams"]
    ) -> "OnrampSession":
        """
        Creates a CryptoOnrampSession object.

        After the CryptoOnrampSession is created, display the onramp session modal using the client_secret.

        Related guide: [Set up an onramp integration](https://docs.stripe.com/docs/crypto/integrate-the-onramp)
        """
        return cast(
            "OnrampSession",
            cls._static_request(
                "post",
                cls.class_url(),
                params=params,
            ),
        )

    @classmethod
    async def create_async(
        cls, **params: Unpack["OnrampSessionCreateParams"]
    ) -> "OnrampSession":
        """
        Creates a CryptoOnrampSession object.

        After the CryptoOnrampSession is created, display the onramp session modal using the client_secret.

        Related guide: [Set up an onramp integration](https://docs.stripe.com/docs/crypto/integrate-the-onramp)
        """
        return cast(
            "OnrampSession",
            await cls._static_request_async(
                "post",
                cls.class_url(),
                params=params,
            ),
        )

    @classmethod
    def list(
        cls, **params: Unpack["OnrampSessionListParams"]
    ) -> ListObject["OnrampSession"]:
        """
        Returns a list of onramp sessions that match the filter criteria. The onramp sessions are returned in sorted order, with the most recent onramp sessions appearing first.
        """
        result = cls._static_request(
            "get",
            cls.class_url(),
            params=params,
        )
        if not isinstance(result, ListObject):
            raise TypeError(
                "Expected list object from API, got %s"
                % (type(result).__name__)
            )

        return result

    @classmethod
    async def list_async(
        cls, **params: Unpack["OnrampSessionListParams"]
    ) -> ListObject["OnrampSession"]:
        """
        Returns a list of onramp sessions that match the filter criteria. The onramp sessions are returned in sorted order, with the most recent onramp sessions appearing first.
        """
        result = await cls._static_request_async(
            "get",
            cls.class_url(),
            params=params,
        )
        if not isinstance(result, ListObject):
            raise TypeError(
                "Expected list object from API, got %s"
                % (type(result).__name__)
            )

        return result

    @classmethod
    def _cls_quote(
        cls, id: str, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        return cast(
            "OnrampSession",
            cls._static_request(
                "post",
                "/v1/crypto/onramp_sessions/{id}/quote".format(
                    id=sanitize_id(id)
                ),
                params=params,
            ),
        )

    @overload
    @staticmethod
    def quote(
        id: str, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        ...

    @overload
    def quote(
        self, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        ...

    @class_method_variant("_cls_quote")
    def quote(  # pyright: ignore[reportGeneralTypeIssues]
        self, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        return cast(
            "OnrampSession",
            self._request(
                "post",
                "/v1/crypto/onramp_sessions/{id}/quote".format(
                    id=sanitize_id(self._data.get("id"))
                ),
                params=params,
            ),
        )

    @classmethod
    async def _cls_quote_async(
        cls, id: str, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        return cast(
            "OnrampSession",
            await cls._static_request_async(
                "post",
                "/v1/crypto/onramp_sessions/{id}/quote".format(
                    id=sanitize_id(id)
                ),
                params=params,
            ),
        )

    @overload
    @staticmethod
    async def quote_async(
        id: str, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        ...

    @overload
    async def quote_async(
        self, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        ...

    @class_method_variant("_cls_quote_async")
    async def quote_async(  # pyright: ignore[reportGeneralTypeIssues]
        self, **params: Unpack["OnrampSessionQuoteParams"]
    ) -> "OnrampSession":
        """
        Refreshes an executable quote for a CryptoOnrampSession.
        """
        return cast(
            "OnrampSession",
            await self._request_async(
                "post",
                "/v1/crypto/onramp_sessions/{id}/quote".format(
                    id=sanitize_id(self._data.get("id"))
                ),
                params=params,
            ),
        )

    @classmethod
    def retrieve(
        cls, id: str, **params: Unpack["OnrampSessionRetrieveParams"]
    ) -> "OnrampSession":
        """
        Retrieves the details of a CryptoOnrampSession that was previously created.
        """
        instance = cls(id, **params)
        instance.refresh()
        return instance

    @classmethod
    async def retrieve_async(
        cls, id: str, **params: Unpack["OnrampSessionRetrieveParams"]
    ) -> "OnrampSession":
        """
        Retrieves the details of a CryptoOnrampSession that was previously created.
        """
        instance = cls(id, **params)
        await instance.refresh_async()
        return instance

    _inner_class_types = {"transaction_details": TransactionDetails}



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._api_resource import APIResource
from stripe._expandable_field import ExpandableField
from stripe._stripe_object import StripeObject
from typing import ClassVar, Optional
from typing_extensions import Literal, Unpack, TYPE_CHECKING

if TYPE_CHECKING:
    from stripe._account import Account
    from stripe._customer import Customer
    from stripe.financial_connections._institution import Institution
    from stripe.params.financial_connections._authorization_retrieve_params import (
        AuthorizationRetrieveParams,
    )


class Authorization(APIResource["Authorization"]):
    """
    An Authorization represents the set of credentials used to connect a group of Financial Connections Accounts.
    """

    OBJECT_NAME: ClassVar[Literal["financial_connections.authorization"]] = (
        "financial_connections.authorization"
    )

    class AccountHolder(StripeObject):
        account: Optional[ExpandableField["Account"]]
        """
        The ID of the Stripe account that this account belongs to. Only available when `account_holder.type` is `account`.
        """
        customer: Optional[ExpandableField["Customer"]]
        """
        The ID for an Account representing a customer that this account belongs to. Only available when `account_holder.type` is `customer`.
        """
        customer_account: Optional[str]
        type: Literal["account", "customer"]
        """
        Type of account holder that this account belongs to.
        """

    class StatusDetails(StripeObject):
        class Active(StripeObject):
            action: Literal["none", "relink_required"]
            """
            The action (if any) to proactively relink the Authorization.
            """
            expected_deactivation_date: int
            """
            When the Authorization is expected to become inactive, if applicable.
            """

        class Inactive(StripeObject):
            action: Literal["none", "relink_required"]
            """
            The action (if any) to relink the inactive Authorization.
            """

        active: Optional[Active]
        inactive: Optional[Inactive]
        _inner_class_types = {"active": Active, "inactive": Inactive}

    account_holder: Optional[AccountHolder]
    """
    The account holder that this authorization belongs to.
    """
    id: str
    """
    Unique identifier for the object.
    """
    institution: Optional[ExpandableField["Institution"]]
    """
    The ID of the Financial Connections Institution this account belongs to. Note that this relationship may sometimes change in rare circumstances (e.g. institution mergers).
    """
    institution_name: str
    """
    The name of the institution that this authorization belongs to.
    """
    livemode: bool
    """
    If the object exists in live mode, the value is `true`. If the object exists in test mode, the value is `false`.
    """
    object: Literal["financial_connections.authorization"]
    """
    String representing the object's type. Objects of the same type share the same value.
    """
    status: Literal["active", "disconnected", "inactive"]
    """
    The status of the connection to the Authorization.
    """
    status_details: StatusDetails

    @classmethod
    def retrieve(
        cls, id: str, **params: Unpack["AuthorizationRetrieveParams"]
    ) -> "Authorization":
        """
        Retrieves the details of an Financial Connections Authorization.
        """
        instance = cls(id, **params)
        instance.refresh()
        return instance

    @classmethod
    async def retrieve_async(
        cls, id: str, **params: Unpack["AuthorizationRetrieveParams"]
    ) -> "Authorization":
        """
        Retrieves the details of an Financial Connections Authorization.
        """
        instance = cls(id, **params)
        await instance.refresh_async()
        return instance

    _inner_class_types = {
        "account_holder": AccountHolder,
        "status_details": StatusDetails,
    }



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from stripe._request_options import RequestOptions
from stripe._stripe_object import UntypedStripeObject
from typing import Dict, List
from typing_extensions import Literal, NotRequired, TypedDict


class PaymentAttemptRecordReportAuthorizedParams(RequestOptions):
    authorized_at: NotRequired[int]
    """
    When the reported payment was authorized. Measured in seconds since the Unix epoch.
    """
    expand: NotRequired[List[str]]
    """
    Specifies which fields in the response should be expanded.
    """
    metadata: NotRequired[
        "Literal['']|Dict[str, str]|UntypedStripeObject[str]"
    ]
    """
    Set of [key-value pairs](https://docs.stripe.com/api/metadata) that you can attach to an object. This can be useful for storing additional information about the object in a structured format. Individual keys can be unset by posting an empty value to them. All keys can be unset by posting an empty value to `metadata`.
    """
    processor_details: NotRequired[
        "PaymentAttemptRecordReportAuthorizedParamsProcessorDetails"
    ]
    """
    Processor information for this payment.
    """


class PaymentAttemptRecordReportAuthorizedParamsProcessorDetails(TypedDict):
    custom: NotRequired[
        "PaymentAttemptRecordReportAuthorizedParamsProcessorDetailsCustom"
    ]
    """
    Information about the custom processor used to make this payment.
    """
    type: Literal["custom"]
    """
    The type of the processor details. An additional hash is included on processor_details with a name matching this value. It contains additional information specific to the processor.
    """


class PaymentAttemptRecordReportAuthorizedParamsProcessorDetailsCustom(
    TypedDict,
):
    payment_reference: str
    """
    An opaque string for manual reconciliation of this payment, for example a check number or a payment processor ID.
    """
